# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Generalized geolocation calculations for geostationary satellites."""

import os
from datetime import datetime
import logging
import numpy as np
from pathlib import Path
from pyresample import utils
from pyresample.geometry import SwathDefinition
from pyresample.kd_tree import get_neighbour_info  # , get_sample_from_neighbour_info
import time
import zarr

from geoips.errors import CoverageError
from geoips.filenames.base_paths import PATHS as gpaths
from geoips.utils.context_managers import import_optional_dependencies

LOG = logging.getLogger(__name__)

# interface = None indicates to the GeoIPS interfaces that this is not a valid
# plugin, and this module will not be added to the GeoIPS plugin registry.
# This allows including python modules within the geoips/plugins directory
# that provide helper or utility functions to the geoips plugins, but are
# not full GeoIPS plugins on their own.
interface = None

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import numexpr as ne


nprocs = 6

try:
    ne.set_num_threads(nprocs)
except Exception:
    LOG.info(
        "Failed numexpr.set_num_threads in %s. If numexpr is not installed and you "
        "need it, install it.",
        __file__,
    )

DONT_AUTOGEN_GEOLOCATION = False
if os.getenv("DONT_AUTOGEN_GEOLOCATION"):
    DONT_AUTOGEN_GEOLOCATION = True

STATIC_GEOLOCDIR = gpaths["GEOIPS_DATA_CACHE_DIR_LONGTERM_GEOLOCATION_STATIC"]

# default dynamic geoloc dir for NRL
DYNAMIC_GEOLOCDIR = gpaths["GEOIPS_DATA_CACHE_DIR_LONGTERM_GEOLOCATION_DYNAMIC"]

READ_GEOLOCDIRS = []
if os.getenv("READ_GEOLOCDIRS"):
    READ_GEOLOCDIRS = os.getenv("READ_GEOLOCDIRS").split(":")


class AutoGenError(Exception):
    """Raise exception on auto generated geolocation error."""

    pass


class CachedGeolocationIndexError(IndexError):
    """Raise exception on cached geolocation IndexError."""

    pass


class CacheNotFoundError(FileNotFoundError):
    """Raise exception if cached data is not found."""

    pass


def check_geolocation_cache_backend(
    cache_backend, supported_backends=("memmap", "zarr")
):
    """Check if requested geolocation cache backend is supported.

    Perhaps this should be converted to a decorator later on?

    Parameters
    ----------
    cache_backend : str
        Library name used to create cached geolocation files (e.g. zarr or memmap)
    supported_backends : list or tuple, optional
        Supported cache backends, by default ("memmap", "zarr")

    Raises
    ------
    ValueError
        If cache_backend is not in supported_backends
    """
    if cache_backend not in supported_backends:
        raise ValueError(
            f"Unsupported cache backend: {cache_backend}."
            f" Supported cache backends: {supported_backends}"
        )


def construct_cache_filename(
    pref, metadata, area_def=None, cache_backend="memmap", chunk_size=None
):
    """Construct a cached file name.

    Parameters
    ----------
    pref : str
        Prefix to identify type of cached data
    metadata : dict
        Top level metadata for dataset
    area_def : pyresample.area_definition, optional
        Area definition subsector of data, by default None
    cache_backend : str, optional
        Specify to use either numpy.memmao or zarray for cache, by default "memmap"
    chunk_size : int, optional
        zarray cache chunk size, by default None

    Returns
    -------
    str
        File name for cached data
    """
    check_geolocation_cache_backend(cache_backend)
    # In order to ensure consistency here, take a sha1 hash of the string representation
    #  of the dictionary values. hash is applied to the object itself, which appears to
    # not be consistent from one Python 3 run to the next.
    # The dictionary values themselves (sorted) SHOULD be consistent between runs.
    from hashlib import sha1

    metadata_string = ""
    for mkey in sorted(metadata.keys()):
        if mkey == "start_datetime":
            continue
        metadata_string += str(metadata[mkey])
    md_hash = sha1(metadata_string.encode("ascii"), usedforsecurity=False).hexdigest()
    # md_hash = hash(frozenset((k, v) for k, v in metadata.items()
    #                                                      if isinstance(v, Hashable)))

    fname = "{}_{}_{}x{}".format(
        pref,
        metadata["scene"],
        metadata["num_lines"],
        metadata["num_samples"],
    )
    if cache_backend != "memmap":
        # Only replace whitespaces with a hyphen when not memmap to prevent regenerating
        # any cached geolocation files
        fname = fname.replace(" ", "-")

    if chunk_size and cache_backend == "zarr":
        # If chunking enabled, include the size in the file name
        fname += f"_chunk{chunk_size}"

    # If the filename format needs to change for the pre-generated geolocation
    # files, please discuss prior to changing.  It will force recreation of all
    # files, which can be problematic for large numbers of sectors

    if area_def:
        ad = area_def
        LOG.info("Using area_definition information ")
        LOG.info(
            "    Using area_definition information for hash: "
            + str(ad.proj_dict.items())
        )
        # sector_hash = hash(frozenset(ad.proj_dict.items()))
        from hashlib import sha1

        metadata_string = ""
        for mkey in sorted(ad.proj_dict.keys()):
            # We don't want to include start/end datetimes in metadata hash!!
            if mkey == "start_datetime":
                continue
            metadata_string += str(ad.proj_dict[mkey])
        sector_hash = sha1(
            metadata_string.encode("ascii"), usedforsecurity=False
        ).hexdigest()
        sect_nlines = ad.shape[0]
        sect_nsamples = ad.shape[1]
        sect_clat = area_def.proj_dict["lat_0"]
        sect_clon = area_def.proj_dict["lon_0"]
        fname += "_{}_{}x{}_{}x{}".format(
            area_def.area_id, sect_nlines, sect_nsamples, sect_clat, sect_clon
        )
        fname += "_{}_{}".format(md_hash, sector_hash)
    else:
        fname += "_{}".format(md_hash)

    if cache_backend == "memmap":
        fname += ".dat"
    elif cache_backend == "zarr":
        fname += ".zarr"
    return fname


def get_data_cache_filename(
    pref,
    metadata,
    area_def=None,
    cache_backend="memmap",
    chunk_size=None,
    scan_datetime=None,
):
    """Get the full file path for a cached calibrated data file.

    Parameters
    ----------
    pref : str
        Prefix to identify type of cached data
    metadata : dict
        Top level metadata for dataset
    area_def : pyresample.area_definition, optional
        Area definition subsector of data, by default None
    cache_backend : str, optional
        Specify to use either numpy.memmao or zarray for cache, by default "memmap"
    chunk_size : int, optional
        zarray cache chunk size, by default None

    Returns
    -------
    str
       Full file path for cached data
    """
    cache_dir = os.path.join(
        gpaths["GEOIPS_DATA_CACHE_DIR_SHORTTERM_CALIBRATED_DATA"],
    )
    if scan_datetime is not None:
        pref += f"_{scan_datetime.strftime('%Y%m%dT%H%M%S.%fZ')}"
    if not os.path.isdir(cache_dir):
        try:
            os.makedirs(cache_dir)
        except FileExistsError:
            pass
    fname = construct_cache_filename(
        pref,
        metadata,
        area_def=area_def,
        cache_backend=cache_backend,
        chunk_size=chunk_size,
    )
    return os.path.join(cache_dir, metadata["platform_name"], fname)


def get_geolocation_cache_filename(
    pref,
    metadata,
    area_def=None,
    geolocation_cache_backend="memmap",
    chunk_size=None,
    solar_angles=False,
):
    """Set the location and filename format for the cached geolocation files.

    There is a separate filename format for satellite latlons and sector latlons

    Notes
    -----
    Changing geolocation filename format will force recreation of all
    files, which can be problematic for large numbers of sectors.
    """
    if solar_angles:
        cache = os.path.join(
            gpaths["GEOIPS_DATA_CACHE_DIR_SHORTTERM_GEOLOCATION_SOLAR_ANGLES"],
            metadata["platform_name"],
        )
    else:
        cache = os.path.join(STATIC_GEOLOCDIR, metadata["platform_name"])
        from geoips.sector_utils.utils import is_dynamic_sector

        if is_dynamic_sector(area_def):
            cache = os.path.join(DYNAMIC_GEOLOCDIR, metadata["platform_name"])
    if not os.path.isdir(cache):
        try:
            os.makedirs(cache)
        except FileExistsError:
            pass

    fname = construct_cache_filename(
        pref,
        metadata,
        area_def=area_def,
        cache_backend=geolocation_cache_backend,
        chunk_size=chunk_size,
    )

    # Check alternative read-only directories (i.e. operational)
    for dirname in READ_GEOLOCDIRS:
        if os.path.exists(os.path.join(dirname, metadata["platform_name"], fname)):
            return os.path.join(dirname, metadata["platform_name"], fname)

    # If not found, return the normal cached filename
    return os.path.join(cache, fname)


def check_for_partial_cache(
    cache_filename, partial_cache_filename, cache_timeout_seconds=30
):
    """Check if a partial data cache exists.

    These partial (temporary) caches are created when the cache is being created, and is
    renamed once complete. If a partial cache is found, try for 30 seconds (default) to
    see if it is renamed to the final cache_filename

    Parameters
    ----------
    cache_filename : str
        Full path to complete cache
    partial_cache_filename : str
        Full path to temporary/partial cache
    cache_timeout_seconds : int, optional
        Retry window to check for final cache if a partial cache is found, by default 30

    Raises
    ------
    CacheNotFoundError
        If cache_filename is not found after retry window
    """
    no_coverage_cache = partial_cache_filename.replace("partial", "no_coverage")
    if Path(partial_cache_filename).exists():
        LOG.info("Found partial cache: %s", partial_cache_filename)
        LOG.info("Will attempt to re-check for complete cache")
        start_time = datetime.now()
        partial_exists = Path(partial_cache_filename).exists()
        while partial_exists:
            elapsed_seconds = (datetime.now() - start_time).total_seconds()
            if elapsed_seconds < cache_timeout_seconds:
                LOG.debug("Partial cache still exists, will check again in 1 second")
                time.sleep(1)
            else:
                LOG.error(
                    "Exceeded cache timeout of %s seconds."
                    " If problem persists, delete %s",
                    cache_timeout_seconds,
                    partial_cache_filename,
                )
                partial_exists = False
    if Path(no_coverage_cache).exists():
        # This will potentially cause an issue specifically for dynamic sectors with
        # interpolated positions where the initial position (ie, storm location at 0Z)
        # is outside satellite coverage, but a later interpolated position (e.g.,
        # interpolated location at 5Z) may have coverage. We are going to accept this
        # case, because the interpolated location will likely still be at edge of scan.
        # Per images below, for TCs in particular, I really don't think we care about
        # this, since they don't move much in 6h and I can't imagine we wouldn't want
        # something at 0Z, but then would want it at 5Z.

        # Note there may be future cases where we would not want to accept this risk, if
        # a dynamic sector moves very quickly, and intermediate interpolated locations
        # could vary drastically from the reported positions at specific times.
        LOG.info("Cache flagged as having no coverage: %s", no_coverage_cache)
    if not Path(cache_filename).exists():
        raise CacheNotFoundError("Cache does not exist: %s", cache_filename)


def create_empty_partial_cache(partial_cache_filename, cache_backend):
    """Create an empty cache with temporary name to denote cache is being processed.

    Supported backends include zarray and memmap.

    Parameters
    ----------
    partial_cache_filename : str
        Full path to temporary/partial cache
    cache_backend : str
        Backend format of cache
    """
    LOG.debug("Creating %s", partial_cache_filename)
    if cache_backend == "zarr":
        # Zarrray data store is a directory
        Path(partial_cache_filename).mkdir(parents=True, exist_ok=True)
    else:
        Path(partial_cache_filename).touch(exist_ok=True)


def rename_partial_cache(partial_cache_filename, final_cache_filename):
    """Rename partial cache to final cache file name.

    Parameters
    ----------
    partial_cache_filename : str
        Full path to temporary/partial cache
    final_cache_filename : str
        Full path to complete cache
    """
    LOG.info("Renaming %s as %s", partial_cache_filename, final_cache_filename)
    Path(partial_cache_filename).rename(Path(final_cache_filename))


def get_geolocation(
    dt,
    gmd,
    fldk_lats,
    fldk_lons,
    BADVALS,
    area_def=None,
    resolution=None,
    geolocation_cache_backend="memmap",
    chunk_size=None,
    cache_solar_angles=False,
    scan_datetime=None,
    resource_tracker=None,
    cache_timeout_seconds=30,
):
    """
    Gather and return the geolocation data for the input metadata.

    Input metadata should be the metadata for a single ABI data file.

    If latitude/longitude have not been calculated with the metadata form the
    input data file they will be recalculated and stored for future use.
    They shouldn't change often. This will be slow the first time it is called
    after a metadata update, but fast thereafter.

    The same is true for satellite zenith and azimuth angles.

    Solar zenith ang azimuth angles are always calculated on the fly.
    This is because they actually change.
    This may be slow for full-disk images.
    """
    check_geolocation_cache_backend(geolocation_cache_backend)
    adname = "None"
    if area_def:
        adname = area_def.area_id

    if resolution:
        adname += f"_{resolution}"

    if resource_tracker is not None:
        key = f"GETGEO: {adname}".replace("None", "ALL")
        resource_tracker.track_resource_usage(
            logstr="MEMUSG", verbose=False, key=key, increment_key=True
        )

    try:
        fldk_sat_zen, fldk_sat_azm = get_satellite_angles(
            gmd,
            fldk_lats,
            fldk_lons,
            BADVALS,
            area_def,
            geolocation_cache_backend=geolocation_cache_backend,
            chunk_size=chunk_size,
            resource_tracker=resource_tracker,
            cache_timeout_seconds=cache_timeout_seconds,
        )
    except AutoGenError:
        return False

    # Determine which indicies will be needed for the input sector if there is one.
    if area_def is not None:
        try:
            lines, samples = get_indexes(
                gmd,
                fldk_lats,
                fldk_lons,
                area_def,
                geolocation_cache_backend=geolocation_cache_backend,
                chunk_size=chunk_size,
                resource_tracker=resource_tracker,
                cache_timeout_seconds=cache_timeout_seconds,
            )
        except AutoGenError:
            return False

        # Get lats, lons, and satellite zenith and azimuth angles for the required
        # points. This may not be entirely appropriate, especially if we want to do
        # something better than nearest neighbor interpolation.
        shape = area_def.shape

        # I'm not entirely sure why this happens from time to time, but sometimes the
        # lines/samples are the correct size, but not the correct shape. I should really
        # figure out the reason for this issue, but it's not entirely clear so in the
        # mean time reshape here as needed.
        if len(lines.shape) == 1 and lines.size == shape[0] * shape[1]:
            LOG.warning("Reshaping lines and samples - should not have to do this...")
            lines = np.reshape(lines, shape)
            samples = np.reshape(samples, shape)

        index_mask = lines != -999

        lons = np.full(shape, -999.1)
        lats = np.full(shape, -999.1)
        sat_zen = np.full(shape, -999.1)
        sat_azm = np.full(shape, -999.1)

        try:
            LOG.info("GETGEO Pulling lons from inds for %s", adname)
            lons[index_mask] = fldk_lons[lines[index_mask], samples[index_mask]]
            LOG.info("GETGEO Pulling lats from inds for %s", adname)
            lats[index_mask] = fldk_lats[lines[index_mask], samples[index_mask]]
            LOG.info("GETGEO Pulling sat_zen from inds for %s", adname)
            sat_zen[index_mask] = fldk_sat_zen[lines[index_mask], samples[index_mask]]
            LOG.info("GETGEO Pulling sat_azm from inds for %s", adname)
            sat_azm[index_mask] = fldk_sat_azm[lines[index_mask], samples[index_mask]]
        except IndexError as resp:
            raise CachedGeolocationIndexError(resp)

    else:
        lats = fldk_lats
        lons = fldk_lons
        sat_zen = fldk_sat_zen
        sat_azm = fldk_sat_azm

    # Get generator for solar zenith and azimuth angles
    LOG.info("GETGEO Must calculate solar zen/azm for sector %s", adname)
    sun_zen, sun_azm = calculate_solar_angles(
        gmd,
        lats,
        lons,
        dt,
        resource_tracker=resource_tracker,
        area_def=area_def,
        geolocation_cache_backend=geolocation_cache_backend,
        chunk_size=chunk_size,
        cache_solar_angles=cache_solar_angles,
        scan_datetime=scan_datetime,
        cache_timeout_seconds=cache_timeout_seconds,
    )
    LOG.info("GETGEO Done calculating solar zen/azm for sector %s", adname)

    # Satellite zenith angle masks are set appropriately for off-disk
    sat_zen = np.ma.masked_less_equal(sat_zen, -999.1)
    sat_azm = np.ma.masked_less_equal(sat_azm, -999.1)

    # Ensure sun_zen and sun_azm are masked appropriately for off-disk values.
    sun_zen = np.ma.masked_where(sat_zen.mask == True, sun_zen)
    sun_azm = np.ma.masked_where(sat_zen.mask == True, sun_azm)

    if resource_tracker is not None:
        resource_tracker.track_resource_usage(
            key=key, checkpoint=True, increment_key=True
        )

    if area_def is not None:
        # area_def.get_lonlats does NOT include any off disk masking.
        lons, lats = area_def.get_lonlats()
        # Set the lats and lons mask to sat_zen, which was set appropriately above.
        lats = np.ma.masked_where(sat_zen.mask == True, lats)
        lons = np.ma.masked_where(sat_zen.mask == True, lons)

    # Make into a dict
    # All the off-disk masking was set above, just catch any straggler -999 values.
    geolocation = {
        "latitude": np.ma.masked_less_equal(lats, -999.1),
        "longitude": np.ma.masked_less_equal(lons, -999.1),
        "satellite_zenith_angle": np.ma.masked_less_equal(sat_zen, -999.1),
        "satellite_azimuth_angle": np.ma.masked_less_equal(sat_azm, -999.1),
        "solar_zenith_angle": np.ma.masked_less_equal(sun_zen, -999.1),
        "solar_azimuth_angle": np.ma.masked_less_equal(sun_azm, -999.1),
    }
    if resource_tracker is not None:
        resource_tracker.track_resource_usage(
            key=key, checkpoint=True, increment_key=True
        )

    try:
        geolocation["Lines"] = np.array(lines)
        geolocation["Samples"] = np.array(samples)
    except NameError:
        pass

    if resource_tracker is not None:
        resource_tracker.track_resource_usage(
            logstr="MEMUSG", verbose=False, key=key, increment_key=True
        )

    return geolocation


def get_satellite_angles(
    metadata,
    lats,
    lons,
    BADVALS,
    sect=None,
    geolocation_cache_backend="memmap",
    chunk_size=None,
    resource_tracker=None,
    cache_timeout_seconds=30,
):
    """Get satellite angles."""
    # If the filename format needs to change for the pre-generated geolocation
    # files, please discuss prior to changing.  It will force recreation of all
    # files, which can be problematic for large numbers of sectors
    check_geolocation_cache_backend(geolocation_cache_backend)
    fname = get_geolocation_cache_filename(
        "GEOSAT",
        metadata,
        geolocation_cache_backend=geolocation_cache_backend,
        chunk_size=chunk_size,
    )
    fname_partial = fname + ".partial"

    if resource_tracker is not None:
        key = "GEO SAT ANGLES: " + str(Path(fname).name)
        if sect:
            key += f"_{sect.area_id}"
        resource_tracker.track_resource_usage(
            logstr="MEMUSG", verbose=False, key=key, increment_key=True
        )

    if not Path(fname).exists() and not Path(fname_partial).exists():
        # First touch a file with the partial cache name to identify we're creating the
        # cache.
        create_empty_partial_cache(fname_partial, geolocation_cache_backend)
        if sect is not None and DONT_AUTOGEN_GEOLOCATION and "tc2019" not in sect.name:
            msg = (
                "GETGEO Requested NO AUTOGEN GEOLOCATION. "
                + "Could not create sat_file for ad {}: {}"
            ).format(metadata["scene"], fname)
            LOG.error(msg)
            raise AutoGenError(msg)

        LOG.info("Calculating satellite zenith and azimuth angles.")
        pi = np.pi
        deg2rad = pi / 180.0  # NOQA
        rad2deg = 180.0 / pi  # NOQA
        sub_lat = 0.0  # NOQA
        sub_lon = metadata["lon0"]  # NOQA
        alt = metadata["H_m"] / 1000.0  # NOQA
        num_lines = metadata["num_lines"]  # NOQA
        num_samples = metadata["num_samples"]  # NOQA

        # Convert lats / lons to radians from sub point
        LOG.debug("Calculating beta")
        # fmt: off
        beta = ne.evaluate("arccos(cos(deg2rad * (lats - sub_lat)) * cos(deg2rad * (lons - sub_lon)))")  # NOQA
        # fmt: on
        bad = lats == BADVALS["Off_Of_Disk"]
        if resource_tracker is not None:
            resource_tracker.track_resource_usage(
                key=key, checkpoint=True, increment_key=True
            )

        # Calculate satellite zenith angle
        LOG.debug("Calculating satellite zenith angle")
        zen = ne.evaluate("alt * sin(beta) / sqrt(1.808e9 - 5.3725e8 * cos(beta))")
        # Where statements take the place of np.clip(zen, - 1.0, 1.0)
        ne.evaluate(
            "rad2deg * arcsin(where(zen < -1.0, -1.0, where(zen > 1.0, 1.0, zen)))",
            out=zen,
        )
        zen[bad] = BADVALS["Off_Of_Disk"]
        if resource_tracker is not None:
            resource_tracker.track_resource_usage(
                key=key, checkpoint=True, increment_key=True
            )

        # Sat azimuth
        LOG.debug("Calculating satellite azimuth angle")
        azm = ne.evaluate("sin(deg2rad * (lons - sub_lon)) / sin(beta)")
        ne.evaluate(
            "rad2deg * arcsin(where(azm < -1.0, -1.0, where(azm > 1.0, 1.0, azm)))",
            out=azm,
        )
        ne.evaluate("where(lats < sub_lat, 180.0 - azm, azm)", out=azm)
        ne.evaluate("where(azm < 0.0, 360.0 + azm, azm)", out=azm)
        azm[bad] = BADVALS["Off_Of_Disk"]
        if resource_tracker is not None:
            resource_tracker.track_resource_usage(
                key=key, checkpoint=True, increment_key=True
            )

        LOG.info("Done calculating satellite zenith and azimuth angles")

        if geolocation_cache_backend == "memmap":
            LOG.info("Storing to %s", fname)
            with open(fname_partial, "w") as df:
                zen.tofile(df)
                azm.tofile(df)
        elif geolocation_cache_backend == "zarr":
            if chunk_size:
                chunks = (chunk_size, chunk_size)
            else:
                chunks = None

            LOG.info("Storing sat zen/azm to %s (chunks=%s)", fname, chunks)
            # NOTE zarr does NOT have a close method, so you can NOT use with context.
            zf = zarr.open(fname_partial, mode="w")
            # Assume azm and zen shape and dtype are the same
            kwargs = {
                "shape": azm.shape,
                "dtype": azm.dtype,
            }
            # As of Python 3.11, can't pass chunks=None into create_dataset
            if chunks:
                # Chunks must be a tuple of the same shape as array.
                kwargs["chunks"] = tuple([chunk_size] * azm.ndim)
            zf.create_dataset("azm", **kwargs)
            zf.create_dataset("zen", **kwargs)
            zf["azm"][:] = azm
            zf["zen"][:] = zen
        rename_partial_cache(fname_partial, fname)

        # Possible switch to xarray based geolocation files, but we lose memmapping.
        # ds = xarray.Dataset({'zeniths':(['x','y'],zen),'azimuths':(['x','y'],azm)})
        # ds.to_netcdf(fname)
    else:
        check_for_partial_cache(fname, fname_partial, cache_timeout_seconds)
        if geolocation_cache_backend == "memmap":
            # Create a memmap to the lat/lon file
            # Nothing will be read until explicitly requested
            # We are mapping this here so that the lats and lons are available when
            # calculating satlelite angles
            LOG.info(
                "GETGEO memmap to {} : lat/lon file for {}".format(
                    fname, metadata["scene"]
                )
            )

            shape = (metadata["num_lines"], metadata["num_samples"])
            offset = 8 * metadata["num_samples"] * metadata["num_lines"]
            zen = np.memmap(fname, mode="r", dtype=np.float64, offset=0, shape=shape)
            if resource_tracker is not None:
                resource_tracker.track_resource_usage(
                    key=key, checkpoint=True, increment_key=True
                )
            azm = np.memmap(
                fname, mode="r", dtype=np.float64, offset=offset, shape=shape
            )
            if resource_tracker is not None:
                resource_tracker.track_resource_usage(
                    key=key, checkpoint=True, increment_key=True
                )
            # Possible switch to xarray based geolocation files, but we lose memmapping.
            # saved_xarray = xarray.load_dataset(fname)
            # zen = saved_xarray['zeniths'].to_masked_array()
            # azm = saved_xarray['azimuths'].to_masked_array()
        elif geolocation_cache_backend == "zarr":
            LOG.info(
                "GETGEO zarr to {} : azm/zen file for {}".format(
                    fname, metadata["scene"]
                )
            )
            shape = (metadata["num_lines"], metadata["num_samples"])
            # chunk_x, chunk_y = [int(x/100) for x in lats.shape]
            # NOTE zarr does NOT have a close method, so you can NOT use with context.
            zf = zarr.open(fname, mode="r")
            azm = zf["azm"]
            zen = zf["zen"]

    if resource_tracker is not None:
        resource_tracker.track_resource_usage(
            logstr="MEMUSG", verbose=False, key=key, increment_key=True
        )

    return zen, azm


def get_indexes(
    metadata,
    lats,
    lons,
    area_def,
    geolocation_cache_backend="memmap",
    chunk_size=None,
    resource_tracker=None,
    cache_timeout_seconds=30,
):
    """
    Return two 2-D arrays containing the X and Y indexes.

    These are indices that should be used from the raw data for the input
    sector definition.
    """
    # The get_neighbor_info function returns three four arrays:
    #    valid_input_index: a 1D boolean array indicating where the source lats and lons
    #                       are valid values (not masked)
    #    valid_output_index: a 1D boolean array indicating where the sector lats and
    #                        lons are valid values (always true everywhere)
    #    index_array: a 1D array of ints indicating which indicies in the flattened
    #                 inputs should be used to fit the sector lats and lons
    #    distance_array: Distances from the source point for each found point.
    #
    # What we do here is feed our data lats/lons to get_neighbour_info.
    # We then reshape valid_input_index to fit our lats/lons and find the 2D indicies
    #   where the input lats and lons were good.
    # We then subset the "good" indicies with index_array to retrieve the required
    #   indicies for the sector.
    # This is complicated because get_neighbour_info does not report the indicies of the
    #   input data, but instead reports the indicies of the flattened data where
    #   valid_input_index is True

    # Get filename for sector indicies

    # If the filename format needs to change for the pre-generated geolocation
    # files, please discuss prior to changing.  It will force recreation of all
    # files, which can be problematic for large numbers of sectors
    check_geolocation_cache_backend(geolocation_cache_backend)
    fname = get_geolocation_cache_filename(
        "GEOINDS",
        metadata,
        area_def,
        geolocation_cache_backend=geolocation_cache_backend,
        chunk_size=None,
    )
    fname_partial = fname + ".partial"
    if Path(fname_partial.replace("partial", "no_coverage")).exists():
        msg = "NO GOOD DATA AVAILABLE, can not read geostationary dataset"
        LOG.info(msg)
        raise CoverageError(msg)

    if resource_tracker is not None:
        key = "GEOINDS: " + str(Path(fname).name)
        resource_tracker.track_resource_usage(
            logstr="MEMUSG", verbose=False, key=key, increment_key=True
        )

    if not Path(fname).exists() and not Path(fname_partial).exists():
        # First touch a file with the partial cache name to identify we're creating the
        # cache.
        create_empty_partial_cache(fname_partial, geolocation_cache_backend)
        if (
            area_def is not None
            and DONT_AUTOGEN_GEOLOCATION
            and "tc2019" not in area_def.area_id
        ):
            msg = (
                "GETGEO Requested NO AUTOGEN GEOLOCATION. "
                + "Could not create inds_file {} for {}"
            ).format(fname, area_def.area_id)
            LOG.error(msg)
            raise AutoGenError(msg)

        # Allocate the full disk area definition
        LOG.info("    GETGEOINDS Masking longitudes")
        lons = np.ma.masked_less(lons, -999.1)
        LOG.info("    GETGEOINDS Wrapping longitudes, pyresample expects -180 to 180")
        # NOTE: it appears pyresample wrap_longitudes can sometimes return a
        # different dtype than passed in. SwathDefinition will fail if dtype
        # does not match between lats and lons. I believe pyresample
        # wrap_longitudes began returning different dtype after numpy 2.0
        # upgrade. It is probably a bug with pyresample, but this will ensure
        # it does not break our processing.
        lons = utils.wrap_longitudes(lons).astype(lats.dtype)
        LOG.info(
            "    GETGEOINDS Creating full disk swath definition for {}".format(
                area_def.area_id
            )
        )
        fldk_ad = SwathDefinition(
            np.ma.masked_less(lons, -999.1), np.ma.masked_less(lats, -999.1)
        )
        ad = area_def

        # Radius of influence will be 10 times the nominal spatial resolution of the
        #   data in meters
        # This uses the only piece of information available concerning resolution
        # in the metadata
        LOG.info(
            "    GETGEOINDS Calculating radius of influence {}".format(area_def.area_id)
        )
        if "res_km" not in metadata.keys():
            shape = lons.shape
            latres = (
                np.abs(
                    lats[int(shape[0] / 2), int(shape[1] / 2)]
                    - lats[int(shape[0] / 2 + 1), int(shape[1] / 2)]
                )
                * 111.1
                * 1000
            )
            lonres = (
                np.abs(
                    lons[int(shape[0] / 2), int(shape[1] / 2)]
                    - lons[int(shape[0] / 2), int(shape[1] / 2 + 1)]
                )
                * 111.1
                * 1000
            )
            # Use larger of the two values times 10 as ROI for interpolation
            # Would be nice to use something more dynamic to save CPU time here
            # Kind of stuck as long as we use pyresample
            metadata["res_km"] = max(latres, lonres) / 1000.0
        roi = (
            metadata["roi_factor"] * 1000.0 * metadata["res_km"]
        )  # roi_factor * resolution in meters
        LOG.info(
            "    GETGEOINDS Running get_neighbour_info %s roi %s res_km %s "
            "roi_factor %s",
            area_def.area_id,
            roi,
            metadata["res_km"],
            metadata["roi_factor"],
        )
        (
            valid_input_index,
            valid_output_index,
            index_array,
            distance_array,
        ) = get_neighbour_info(
            fldk_ad, ad, radius_of_influence=roi, neighbours=1, nprocs=nprocs
        )
        LOG.info(
            "    GETGEOINDS Getting good lines and samples {}".format(area_def.area_id)
        )
        good_lines, good_samples = np.where(valid_input_index.reshape(lats.shape))
        if len(good_lines) == 0 and len(good_samples) == 0:
            # Rename the partial cache and flag as no coverage if it exists.
            if Path(fname_partial).exists():
                rename_partial_cache(
                    fname_partial, fname_partial.replace("partial", "no_coverage")
                )
            raise CoverageError(
                "NO GOOD DATA AVAILABLE, can not read geostationary dataset"
            )
        LOG.info(
            "    GETGEOINDS Reshaping lines and samples {}".format(area_def.area_id)
        )
        # When get_neighbour_info does not find a good value for a specific location it
        #   fills index_array with the maximum index + 1.  So, just throw away all of
        #   the out of range indexes.
        index_mask = index_array == len(good_lines)
        # good_index_array = index_array[np.where(index_array != len(good_lines))]
        lines = np.empty(ad.size, dtype=np.int64)
        lines[index_mask] = -999.1
        lines[~index_mask] = good_lines[index_array[~index_mask]]
        samples = np.empty(ad.size, dtype=np.int64)
        samples[index_mask] = -999.1
        samples[~index_mask] = good_samples[index_array[~index_mask]]

        LOG.info(
            "    GETGEOINDS Writing to {} : inds_file for {}".format(
                fname, area_def.area_id
            )
        )
        if geolocation_cache_backend == "memmap":
            # Store indicies for sector
            with open(str(fname_partial), "w") as df:
                lines.tofile(df)
                samples.tofile(df)
            # Store indicies for sector
            # Possible switch to xarray based geolocation files, but we lose memmapping.
            # ds = xarray.Dataset({'lines':(['x'],lines),'samples':(['x'],samples)})
            # ds.to_netcdf(fname_partial)
        elif geolocation_cache_backend == "zarr":
            if chunk_size:
                chunks = (chunk_size, chunk_size)
            else:
                chunks = None

            LOG.info("Storing lines/samples to %s (chunks=%s)", fname, chunks)
            # NOTE zarr does NOT have a close method, so you can NOT use with context.
            zf = zarr.open(fname_partial, mode="w")
            # Assume both arrays have the same shape and dtype
            kwargs = {
                "shape": lines.shape,
                "dtype": lines.dtype,
            }
            # As of Python 3.11, can't pass chunks=None into create_dataset
            if chunks:
                # Chunks must be a tuple of the same shape as array.
                kwargs["chunks"] = tuple([chunk_size] * lines.ndim)
            zf.create_dataset("lines", **kwargs)
            zf.create_dataset("samples", **kwargs)
            zf["lines"][:] = lines
            zf["samples"][:] = samples
        rename_partial_cache(fname_partial, fname)

    else:
        check_for_partial_cache(fname, fname_partial, cache_timeout_seconds)
        LOG.info(
            "GETGEO to %s : inds file for %s, roi_factor %s",
            fname,
            metadata["scene"],
            metadata["roi_factor"],
        )
        LOG.info(
            "GETGEO to %s : lat/lon file for %s, roi_factor %s",
            fname,
            metadata["scene"],
            metadata["roi_factor"],
        )
        shape = area_def.shape
        try:
            if geolocation_cache_backend == "memmap":
                # Create a memmap to the lat/lon file
                # Nothing will be read until explicitly requested
                # We are mapping this here so that the lats and lons are available when
                # calculating satlelite angles.
                # LOG.info(
                #   'GETGEO memmap to %s : inds file for %s, roi_factor %s, res_km %s',
                #   fname,
                #   metadata['scene'],
                #   metadata['roi_factor'],
                #   metadata['res_km']
                # )
                # LOG.info(
                #   'GETGEO memmap to %s : %s file for %s, roi_factor %s, res_km %s',
                #    fname,
                #   "lat/lon",
                #   metadata['scene'],
                #   metadata['roi_factor'],
                #   metadata['res_km']
                # )
                offset = 8 * shape[0] * shape[1]
                LOG.info(
                    "GETGEO memmap from %s : lines for %s, shape %s",
                    fname,
                    metadata["scene"],
                    shape,
                )
                lines = np.memmap(
                    fname, mode="r", dtype=np.int64, offset=0, shape=shape
                )
                LOG.info(
                    "GETGEO memmap from %s : samples for %s, offset %s",
                    fname,
                    metadata["scene"],
                    offset,
                )
                samples = np.memmap(
                    fname, mode="r", dtype=np.int64, offset=offset, shape=shape
                )
            elif geolocation_cache_backend == "zarr":
                # Load lines/samples from pre-calculated zarr
                LOG.info(
                    "GETGEO zarr to {} : inds file for {}".format(
                        fname, area_def.area_id
                    )
                )
                # NOTE zarr does NOT have close method, so you can NOT use with context.
                zf = zarr.open(fname, mode="r")
                if "lines" not in zf or "samples" not in zf:
                    LOG.exception(
                        "lines and samples not in zarr directory, corrupt zarr? "
                        f"Please remove {fname} and re-run this script."
                    )
                    raise RuntimeError(
                        "lines and samples not in zarr directory, corrupt zarr? "
                        f"Please remove the directory:\n"
                        f"{fname}\n"
                        "and re-run this script."
                    )
                lines = zf["lines"]
                samples = zf["samples"]
                lines = np.reshape(lines, shape=shape)
                samples = np.reshape(samples, shape=shape)
        except ValueError as resp:
            LOG.warning(
                "Mismatched geolocation file size (Empty?  No coverage?  Or old "
                "sector of different shape?"
            )
            raise CachedGeolocationIndexError(resp)

    if resource_tracker is not None:
        resource_tracker.track_resource_usage(
            logstr="MEMUSG", verbose=False, key=key, increment_key=True
        )

    # Possible switch to xarray based geolocation files, but we lose memmapping.
    # saved_xarray = xarray.load_dataset(fname)
    # lines= saved_xarray['lines'].to_masked_array()
    # samples = saved_xarray['samples'].to_masked_array()
    return lines, samples


def calculate_solar_angles(
    metadata,
    lats,
    lons,
    dt,
    resource_tracker=None,
    area_def=None,
    geolocation_cache_backend=None,
    chunk_size=None,
    cache_solar_angles=False,
    scan_datetime=None,
    cache_timeout_seconds=30,
):
    """Calculate solar angles."""
    # If debug is set to True, memory savings will be turned off in order to keep
    # all calculated results for inspection.
    # If set to False, variables will attempt to reuse memory when possible which
    # will result in some results being overwritten when no longer needed.
    debug = False
    if area_def:
        adname = area_def.area_id
    else:
        adname = "None"
    if cache_solar_angles:
        check_geolocation_cache_backend(geolocation_cache_backend)
        prefix = "GEOSOL"
        if scan_datetime is not None:
            prefix += f"_{scan_datetime.strftime('%Y%m%dT%H%M%S.%fZ')}"
        fname = get_geolocation_cache_filename(
            prefix,
            metadata,
            area_def=area_def,
            geolocation_cache_backend=geolocation_cache_backend,
            chunk_size=chunk_size,
            solar_angles=True,
        )
        fname_partial = fname + ".partial"
        cache_exists = Path(fname).exists()
    else:
        cache_exists = False

    LOG.info("Calculating solar zenith and azimuth angles.")

    if resource_tracker is not None:
        key = f"GEO SOLAR ANGLES: {adname}_{lats.shape[0]}x{lats.shape[1]}".replace(
            "None", "ALL"
        )
        resource_tracker.track_resource_usage(
            logstr="MEMUSG", verbose=False, key=key, increment_key=True
        )

    # Getting good value mask
    # good = lats > -999
    # good_lats = lats[good]
    # good_lons = lons[good]
    if cache_solar_angles is False or (cache_solar_angles and cache_exists is False):
        if cache_solar_angles and not cache_exists and not Path(fname_partial).exists():
            # First touch a file with the partial cache name to identify we're creating
            # the cache.
            create_empty_partial_cache(fname_partial, geolocation_cache_backend)
        # Constants
        pi = np.pi
        pi2 = 2 * pi  # NOQA
        num_lines = metadata["num_lines"]
        num_samples = metadata["num_samples"]
        shape = (num_lines, num_samples)  # NOQA
        size = num_lines * num_samples  # NOQA
        deg2rad = pi / 180.0  # NOQA
        rad2deg = 180.0 / pi  # NOQA

        # Calculate any non-data dependent quantities
        jday = float(dt.strftime("%j"))
        a1 = (1.00554 * jday - 6.28306) * (pi / 180.0)
        a2 = (1.93946 * jday - 23.35089) * (pi / 180.0)
        et = -7.67825 * np.sin(a1) - 10.09176 * np.sin(a2)  # NOQA

        # Solar declination radians
        LOG.debug("Calculating delta")
        delta = deg2rad * 23.4856 * np.sin(np.deg2rad(0.9683 * jday - 78.00878))  # NOQA

        # Pre-generate sin and cos of latitude
        LOG.debug("Calculating sin and cos")
        sin_lat = ne.evaluate("sin(deg2rad * lats)")  # NOQA
        cos_lat = ne.evaluate("cos(deg2rad * lats)")  # NOQA

        # Hour angle
        LOG.debug("Initializing hour angle")
        solar_time = dt.hour + dt.minute / 60.0 + dt.second / 3600.0  # NOQA
        h_ang = ne.evaluate(
            "deg2rad * ((solar_time + lons / 15.0 + et / 60.0 - 12.0) * 15.0)"
        )

        # Pre-allocate all required arrays
        # This avoids having to allocate them again every time the generator is accessed
        LOG.debug("Allocating arrays")
        sun_elev = np.empty_like(h_ang)

        # Hour angle at all points in radians
        LOG.debug("Calculating hour angle")

        # Sun elevation
        LOG.debug("Calculating sun elevation angle using sin and cos")
        ne.evaluate(
            "arcsin(sin_lat * sin(delta) + cos_lat * cos(delta) * cos(h_ang))",
            out=sun_elev,
        )  # NOQA

        LOG.debug("Calculating caz")
        # No longer need sin_lat and this saves 3.7GB
        if not debug:
            caz = sin_lat
        else:
            caz = np.empty_like(sin_lat)
        ne.evaluate(
            "-cos_lat * sin(delta) + sin_lat * cos(delta) * cos(h_ang) / cos(sun_elev)",
            out=caz,
        )  # NOQA

        LOG.debug("Calculating az")
        # No longer need h_ang and this saves 3.7GB
        if not debug:
            az = h_ang
        else:
            az = np.empty_like(h_ang)
        ne.evaluate("cos(delta) * sin(h_ang) / cos(sun_elev)", out=az)  # NOQA
        # No longer need sin_lat and this saves 3.7GB
        if not debug:
            sun_azm = cos_lat
        else:
            sun_azm = np.empty_like(cos_lat)
        ne.evaluate(
            "where(az <= -1, -pi / 2.0, where(az > 1, pi / 2.0, arcsin(az)))",
            out=sun_azm,
        )

        LOG.debug("Calculating solar zenith angle")
        # No longer need sun_elev and this saves 3.7GB RAM
        if not debug:
            sun_zen = sun_elev
        else:
            sun_zen = np.empty_like(sun_elev)
        ne.evaluate("90.0 - rad2deg * sun_elev", out=sun_zen)

        LOG.debug("Calculating solar azimuth angle")
        ne.evaluate(
            "where(caz <= 0, pi - sun_azm,where(az <= 0, 2.0 * pi + sun_azm, sun_azm))",
            out=sun_azm,
        )
        sun_azm += pi
        ne.evaluate(
            "where(sun_azm > 2.0 * pi, sun_azm - 2.0 * pi, sun_azm)", out=sun_azm
        )
        # ne.evaluate('where(caz <= 0, pi - sun_azm, sun_azm) + pi', out=sun_azm)
        # ne.evaluate('rad2deg * where(sun_azm < 0, sun_azm + pi2, where(sun_azm >= pi2,
        #                                                     sun_azm - pi2, sun_azm))',
        #                                                                  out=sun_azm)
        if cache_solar_angles:
            if geolocation_cache_backend == "memmap":
                LOG.info("Storing to %s", fname_partial)
                with open(fname_partial, "w") as df:
                    sun_zen.tofile(df)
                    sun_azm.tofile(df)
            elif geolocation_cache_backend == "zarr":
                if chunk_size:
                    chunks = (chunk_size, chunk_size)
                else:
                    chunks = None

                LOG.info("Storing solar angles to %s (chunks=%s)", fname, chunks)
                # NOTE zarr does NOT have close method, so you can NOT use the context.
                zf = zarr.open(fname_partial, mode="w")
                # Assume both arrays have the same shape and dtype
                kwargs = {
                    "shape": sun_azm.shape,
                    "dtype": sun_azm.dtype,
                }
                # As of Python 3.11, can't pass chunks=None into create_dataset
                if chunks:
                    # Chunks must be a tuple of the same shape as array.
                    kwargs["chunks"] = tuple([chunk_size] * sun_azm.ndim)
                zf.create_dataset("sun_azm", **kwargs)
                zf.create_dataset("sun_zen", **kwargs)
                zf["sun_azm"][:] = sun_azm
                zf["sun_zen"][:] = sun_zen
            rename_partial_cache(fname_partial, fname)

        LOG.info("Done calculating solar zenith and azimuth angles")
    else:
        check_for_partial_cache(fname, fname_partial, cache_timeout_seconds)
        if geolocation_cache_backend == "memmap":
            # Create a memmap to the lat/lon file
            # Nothing will be read until explicitly requested
            # We are mapping this here so that the lats and lons are available when
            # calculating satlelite angles
            LOG.info(
                "GETGEO memmap to {} : lat/lon file for {}".format(
                    fname, metadata["scene"]
                )
            )

            shape = (metadata["num_lines"], metadata["num_samples"])
            offset = 8 * metadata["num_samples"] * metadata["num_lines"]
            sun_zen = np.memmap(
                fname, mode="r", dtype=np.float64, offset=0, shape=shape
            )
            if resource_tracker is not None:
                resource_tracker.track_resource_usage(
                    key=key, checkpoint=True, increment_key=True
                )
            sun_azm = np.memmap(
                fname, mode="r", dtype=np.float64, offset=offset, shape=shape
            )
            if resource_tracker is not None:
                resource_tracker.track_resource_usage(
                    key=key, checkpoint=True, increment_key=True
                )
            # Possible switch to xarray based geolocation files, but we lose memmapping.
            # saved_xarray = xarray.load_dataset(fname)
            # zen = saved_xarray['zeniths'].to_masked_array()
            # azm = saved_xarray['azimuths'].to_masked_array()
        elif geolocation_cache_backend == "zarr":
            LOG.info(
                "GETGEO zarr to {} : sun_azm/sun_zen file for {}".format(
                    fname, metadata["scene"]
                )
            )
            shape = (metadata["num_lines"], metadata["num_samples"])
            # chunk_x, chunk_y = [int(x/100) for x in lats.shape]
            # NOTE zarr does NOT have a close method, so you can NOT use with context.
            zf = zarr.open(fname, mode="r")
            sun_azm = zf["sun_azm"]
            sun_zen = zf["sun_zen"]

    if resource_tracker is not None:
        resource_tracker.track_resource_usage(
            logstr="MEMUSG", verbose=False, key=key, increment_key=True
        )

    return sun_zen, sun_azm



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Command line script for kicking off geoips based procflows."""

import argparse
import logging
from os.path import abspath, exists
from os import getenv
from ast import literal_eval
import isodate

LOG = logging.getLogger(__name__)


def check_command_line_args(arglist, argdict):
    """Check formatting of command line arguments.

    Parameters
    ----------
    arglist : list of str
        List of desired command line arguments to check within argdict for
        appropriate formatting
    argdict : dict
        Dictionary of command line arguments

    Returns
    -------
    bool
        Return True if all arguments are of appropriate formatting.

    Raises
    ------
    TypeError
        Incorrect command line formatting
    """
    if arglist is None:
        return True
    if "filenames" in arglist:
        if argdict["filenames"] and not isinstance(argdict["filenames"], list):
            raise TypeError(
                'Must pass list of strings for "filenames" dictionary entry'
            )
        for fname in argdict["filenames"]:
            if not exists(fname):
                raise IOError(
                    f"Filename {fname} does not exist - all requested files must "
                    "exist on disk"
                )
        # LOG.info('COMMANDLINEARG filenames: %s', argdict['filenames'])
    if "self_register_dataset" in arglist:
        if argdict["self_register_dataset"] and not isinstance(
            argdict["self_register_dataset"], str
        ):
            raise TypeError(
                'Must pass string for "self_register_dataset" dictionary entry'
            )
        LOG.info(
            "COMMANDLINEARG self_register_dataset: %s", argdict["self_register_dataset"]
        )
    if "sectored_read" in arglist:
        if argdict["sectored_read"] and not isinstance(argdict["sectored_read"], bool):
            raise TypeError('Must pass bool for "sectored_read" dictionary entry')
        LOG.info("COMMANDLINEARG sectored_read: %s", argdict["sectored_read"])
    if "resampled_read" in arglist:
        if argdict["resampled_read"] and not isinstance(
            argdict["resampled_read"], bool
        ):
            raise TypeError('Must pass bool for "resampled_read" dictionary entry')
        LOG.info("COMMANDLINEARG resampled_read: %s", argdict["resampled_read"])
    if "sector_list" in arglist:
        if argdict["sector_list"] and not isinstance(argdict["sector_list"], list):
            raise TypeError(
                "Must pass list of strings for requested static sector plugins"
            )
        LOG.info("COMMANDLINEARG sector_list: %s", argdict["sector_list"])
    if "tcdb_sector_list" in arglist:
        if argdict["tcdb_sector_list"] and not isinstance(
            argdict["tcdb_sector_list"], list
        ):
            raise TypeError(
                'Must pass list of strings for "tcdb_sector_list" dictionary entry'
            )
        LOG.info("COMMANDLINEARG tcdb_sector_list: %s", argdict["tcdb_sector_list"])
    if "tcdb" in arglist:
        if argdict["tcdb"] not in [True, False]:
            raise TypeError("tcdb dictionary entry must be bool (True or False)")
        LOG.info("COMMANDLINEARG tcdb: %s", argdict["tcdb"])
    if "product_name" in arglist:
        if argdict["procflow"] == "single_source" and not argdict["product_name"]:
            raise TypeError(
                'Must pass a string for "product_name" when running "single_source"'
            )
        if argdict["product_name"] and not isinstance(argdict["product_name"], str):
            raise TypeError(
                'Must pass a single string for "product_name" dictionary entry'
            )
        LOG.info("COMMANDLINEARG product_name: %s", argdict["product_name"])
    if "product_options" in arglist:
        if argdict["product_options"] and not isinstance(
            argdict["product_options"], str
        ):
            raise TypeError('Must pass string for "product_options" dictionary entry')
        LOG.info("COMMANDLINEARG product_options: %s", argdict["product_options"])
    if "reader_name" in arglist:
        if argdict["reader_name"] and not isinstance(argdict["reader_name"], str):
            raise TypeError('Must pass string for "reader_name" dictionary entry')
        LOG.info("COMMANDLINEARG reader_name: %s", argdict["reader_name"])
    if "minimum_coverage" in arglist:
        if argdict["minimum_coverage"] and not isinstance(
            argdict["minimum_coverage"], float
        ):
            raise TypeError('Must pass float for "minimum_coverage" dictionary entry')
        LOG.info("COMMANDLINEARG minimum_coverage: %s", argdict["minimum_coverage"])
    if "output_config" in arglist:
        if argdict["output_config"] and not isinstance(argdict["output_config"], str):
            raise TypeError(
                'Must pass a single string for "output_config" dictionary entry'
            )
        LOG.info("COMMANDLINEARG output_config: %s", argdict["output_config"])
    if "output_file_list_fname" in arglist:
        if argdict["output_file_list_fname"] and not isinstance(
            argdict["output_file_list_fname"], str
        ):
            raise TypeError(
                "Must pass a single string for 'output_file_list_fname' dictionary "
                "entry"
            )
        LOG.info(
            "COMMANDLINEARG output_file_list_fname: %s",
            argdict["output_file_list_fname"],
        )
    if "sector_adjuster" in arglist:
        if argdict["sector_adjuster"] and not isinstance(
            argdict["sector_adjuster"], str
        ):
            raise TypeError(
                'Must pass a single string for "sector_adjuster" dictionary entry'
            )
        LOG.info("COMMANDLINEARG sector_adjuster: %s", argdict["sector_adjuster"])
    if "reader_defined_area_def" in arglist:
        if argdict["reader_defined_area_def"] and not isinstance(
            argdict["reader_defined_area_def"], bool
        ):
            raise TypeError(
                'Must pass True or False for "reader_defined_area_def" dictionary entry'
            )
        LOG.info(
            "COMMANDLINEARG reader_defined_area_def: %s",
            argdict["reader_defined_area_def"],
        )

    return True


def get_argparser(
    arglist=None, description=None, add_args_func=None, check_args_func=None
):
    """Get argparse.ArgumentParser with all standard arguments added.

    Parameters
    ----------
    arglist : list, optional
        list of requested arguments to add to the ArgumentParser, default None.
        if None, include all arguments
    description : str, optional
        String description of arguments, default None
    add_args_func : function, optional
        Alternative "add_args" function, default None
        If None, use internal "add_args"
    check_args_func: function, optional
        Alternative "check_args" function, default None
        If None, use internal "check_args"

    Returns
    -------
    argparse.ArgumentParser
        Including all requested/required command line arguments.
    """
    if add_args_func is None:
        add_args_func = add_args
    if check_args_func is None:
        check_args_func = check_command_line_args
    parser = argparse.ArgumentParser(description=description)
    add_args_func(parser, arglist)
    return parser


def get_command_line_args(
    arglist=None, description=None, add_args_func=None, check_args_func=None
):
    """Parse command line arguments specified by the requested list of arguments.

    Parameters
    ----------
    arglist : list, optional
        list of requested arguments to add to the ArgumentParser, default None.
        if None, include all arguments
    description : str, optional
        String description of arguments, default None
    add_args_func : function, optional
        Alternative "add_args" function, default None
        If None, use internal "add_args"
    check_args_func: function, optional
        Alternative "check_args" function, default None
        If None, use internal "check_args"

    Returns
    -------
    dict
        Dictionary of command line arguments
    """
    parser = get_argparser(arglist, description, add_args_func, check_args_func)
    argdict = parser.parse_args()
    if arglist and "filenames" in arglist and "procflow" in arglist:
        check_args_func(["filenames", "procflow"], argdict.__dict__)
    return argdict


def add_args(parser, arglist=None, legacy=False):
    """List of available standard arguments for calling data processing command line.

    Parameters
    ----------
    parser : ArgumentParser
        argparse ArgumentParser to add appropriate arguments
    arglist : list, optional
        list of requested arguments to add to the ArgumentParser, default None.
        if None, include all arguments
    legacy : bool, optional
        Represents whether or not a legacy 'run_procflow' or 'data_fusion_procflow'
        was called

    Returns
    -------
        No return values (parser modified in place)
    """
    if arglist is None or "filenames" in arglist:
        parser.add_argument(
            "filenames",
            nargs="*",
            default=None,
            type=abspath,
            help="""Fully qualified paths to data files to be processed.""",
        )

    if arglist is None or "outdir" in arglist:
        parser.add_argument(
            "-o",
            "--outdir",
            default=getenv("GEOIPS_OUTDIRS"),
            help="""Path to write output files.  Defaults to GEOIPS_OUTDIRS.""",
        )

    if arglist is None or "logging_level" in arglist:
        parser.add_argument(
            "-l",
            "--logging_level",
            choices=["INTERACTIVE", "INFO", "DEBUG", "WARNING", "ERROR", "CRITICAL"],
            default=None,
            help="""Specify logging config level for GeoIPS commands.""",
            type=str.upper,
        )
    if arglist is None or "rebuild_registries" in arglist:
        parser.add_argument(
            "--rebuild_registries",
            type=bool,
            choices=[True, False],
            # Will default to $GEOIPS_REBUILD_REGISTRIES if set, else True
            default=getenv("GEOIPS_REBUILD_REGISTRIES", True),
            help=(
                "Whether or not you want to rebuild plugin registries if you encounter "
                "a failure when retrieving a plugin."
            ),
        )

    sect_group = parser.add_argument_group(
        title="Sector Requests: General arguments for sectors"
    )
    if arglist is None or "sector_adjuster" in arglist:
        sect_group.add_argument(
            "--sector_adjuster",
            nargs="?",
            default=None,
            help="""Specify sector adjuster to be used within processing, located in:
                            <package>.plugins.modules.sector_adjusters.
                                <myadjuster>.<myadjuster>""",
        )
    if arglist is None or "sector_adjuster_kwargs" in arglist:
        sect_group.add_argument(
            "--sector_adjuster_kwargs",
            nargs="?",
            default={},
            type=literal_eval,
            help="""Specify sector_adjuster kwargs that should be used for
                            this sector_adjuster. Should be formatted as a json
                            dictionary string""",
        )
    if arglist is None or "window_start_time" in arglist:
        sect_group.add_argument(
            "--window_start_time",
            nargs="?",
            default=None,
            type=isodate.parse_datetime,
            help="""If specified, only include data between
                    window_start_time and window_end_time.
                    Must be specified with window_end_time.
                    This option will override any default time ranges determined
                    based on potential dynamic sector times.
                            Defaults to None (include all data).""",
        )
    if arglist is None or "window_end_time" in arglist:
        sect_group.add_argument(
            "--window_end_time",
            nargs="?",
            default=None,
            type=isodate.parse_datetime,
            help="""If specified, only include data between
                    window_start_time and window_end_time.
                    Must be specified with window_start_time.
                    This option will override any default time ranges determined
                    based on potential dynamic sector times.
                            Defaults to None (include all data).""",
        )

    tc_group = parser.add_argument_group(
        title="Sector Requests: General arguments for TC sectors"
    )
    if arglist is None or "tc_spec_template" in arglist:
        tc_group.add_argument(
            "--tc_spec_template",
            nargs="?",
            default=None,
            help="""YAML plugin for creating appropriate TC sector using
                    shape/resolution from current storm location.""",
        )

    trackfile_group = parser.add_argument_group(
        title="Sector Requests: TC trackfile-based sectors"
    )
    if arglist is None or "trackfiles" in arglist:
        trackfile_group.add_argument(
            "--trackfiles",
            nargs="*",
            default=None,
            help="""Specify TC trackfiles to include in processing
                            If --trackfile_sector_list is included,
                                limit to the storms in list
                            If --trackfile_sector_list is not included,
                                process all storms""",
        )
    if arglist is None or "trackfile_parser" in arglist:
        trackfile_group.add_argument(
            "--trackfile_parser",
            nargs="?",
            default=None,
            help="""Specify TC trackfile parser to use with trackfiles, located in:
                            geoips*.plugins.modules.sector_metadata_generators .
                                myparsername.myparsername,
                            The trackfile_parser string should be the parser module
                            name (no .py)""",
        )
    if arglist is None or "trackfile_sector_list" in arglist:
        trackfile_group.add_argument(
            "--trackfile_sector_list",
            nargs="*",
            default=None,
            help="""A list of sector names found specified trackfiles to include
                    in processing. Of format: tc2020io01amphan""",
        )

    tcdb_group = parser.add_argument_group(title="Sector Requests: TC tracks database")
    if arglist is None or "tcdb_sector_list" in arglist:
        tcdb_group.add_argument(
            "--tcdb_sector_list",
            nargs="*",
            default=None,
            help="""A list of sector names found in tc database to include in
                    processing. Of format: tc2020io01amphan""",
        )
    if arglist is None or "tcdb" in arglist:
        tcdb_group.add_argument(
            "--tcdb",
            action="store_true",
            help="""Call with --tcdb to include the matching TC database sectors
                            within processing
                            If --tcdb_sector_list is also included,
                                limit the storms to those in list
                            If --tcdb_sector_list is not included,
                                process all matching storms.""",
        )

    static_group = parser.add_argument_group(
        title="Sector Requests: Static YAML sectorfiles"
    )
    if arglist is None or "sectored_read" in arglist:
        static_group.add_argument(
            "--sectored_read",
            action="store_true",
            help="""Call with --sectored_read to specify to sector the data
                            to specified area_defs during reading (ie, do not
                            read all data into memory in advance).""",
        )
    if arglist is None or "resampled_read" in arglist:
        static_group.add_argument(
            "--resampled_read",
            action="store_true",
            help="""Call with --resampled_read to specify to resample the data
                            to specified area_defs during reading (ie, do not
                            read all data into memory in advance).
                            This is required only for some geostationary readers""",
        )
    if arglist is None or "self_register_dataset" in arglist:
        static_group.add_argument(
            "--self_register_dataset",
            nargs="?",
            default=None,
            help="""Specify to register output data to the dataset specified by
                            self_register_dataset option.""",
        )
    if arglist is None or "self_register_source" in arglist:
        static_group.add_argument(
            "--self_register_source",
            nargs="?",
            default=None,
            help="""Specify to register output data to the dataset specified by
                            self_register_dataset / self_register_source options.""",
        )
    if arglist is None or "reader_defined_area_def" in arglist:
        static_group.add_argument(
            "--reader_defined_area_def",
            action="store_true",
            help="""Call with --reader_defined_area_def to specify to use only
                            area_definition defined within the reader. This option
                            supercedes all other sector-specifying options.""",
        )
    if arglist is None or "sector_list" in arglist:
        static_group.add_argument(
            "-s",
            "--sector_list",
            nargs="*",
            default=None,
            help="""A list of short sector plugin names found within YAML sectorfiles
                    over which the data file should be processed.""",
        )

    prod_group = parser.add_argument_group(title="Product specification options")
    if arglist is None or "product_name" in arglist:
        prod_group.add_argument(
            "--product_name",
            nargs="?",
            default=None,
            help="""Name of product to produce.""",
        )
    if arglist is None or "minimum_coverage" in arglist:
        prod_group.add_argument(
            "--minimum_coverage",
            nargs="?",
            default=None,
            type=float,
            help="""Minimum percent coverage required to produce product.
                            Defaults to 10.""",
        )
    if arglist is None or "product_options" in arglist:
        prod_group.add_argument(
            "--product_options",
            nargs="?",
            default=None,
            help="""Specify product specific options (these must be parsed
                            within the individual product scripts)""",
        )
    if arglist is None or "product_spec_override" in arglist:
        prod_group.add_argument(
            "--product_spec_override",
            nargs="?",
            default={},
            type=literal_eval,
            help="""Specify product spec fields to override the default specifications.
                            Should be formatted as a json dictionary string""",
        )

    comp_group = parser.add_argument_group(
        title="Options for specifying output comparison"
    )
    if arglist is None or "output_file_list_fname" in arglist:
        comp_group.add_argument(
            "--output_file_list_fname",
            nargs="?",
            default=None,
            help="""Specify full path to a single file to store the complete list
                            of output files produced during the current run""",
        )

    if arglist is None or "compare_path" in arglist:
        comp_group.add_argument(
            "--compare_path",
            nargs="?",
            default=None,
            help="""Specify full path to single directory
                            (with <product> and <output> wildcards)
                            to be used for comparisons for ALL current outputs.""",
        )

    if arglist is None or "compare_paths_override" in arglist:
        comp_group.add_argument(
            "--compare_paths_override",
            nargs="?",
            default={},
            type=literal_eval,
            help="""NOT YET IMPLEMENTED Specify dictionary of full paths to directories
                            (with <product> and <output> wildcards) containing
                            output products to compare with current outputs.
                            This should be formatted as a json dictionary string, with
                            YAML output config output_types as keys, and full directory
                            comparison output path as values.  Special key "all" will
                            pertain to all output types.""",
        )

    procflow_group = parser.add_argument_group(
        title="Processing workflow specifications"
    )
    if arglist is None or "procflow" in arglist:
        if legacy:
            help_str = (
                "Specify procflow that should be followed for this file, located in "
                "geoips.plugins.modules.procflows.myprocflowname.name. The procflow "
                "string should be the procflow module file name (excluding '.py' )."
            )
        else:
            help_str = argparse.SUPPRESS
        procflow_group.add_argument(
            "--procflow",
            default=None,
            help=help_str,
        )

    if arglist is None or "filename_formatter" in arglist:
        procflow_group.add_argument(
            "--filename_formatter",
            nargs="?",
            default="geoips_fname",
            help="""Specify filename format module_name that should be used for
                            this file, where each filename_module_name is
                            'myfilemodule' where:
                                from geoips*.filenames.myfilemodule import myfilemodule
                            would be the appropriate import statement""",
        )
    if arglist is None or "filename_formatter_kwargs" in arglist:
        procflow_group.add_argument(
            "--filename_formatter_kwargs",
            nargs="?",
            default={},
            type=literal_eval,
            help="""Specify filename format kwargs that should be used for
                            this filename_formatter. Should be formatted as a json
                            dictionary string""",
        )

    if arglist is None or "metadata_filename_formatter" in arglist:
        procflow_group.add_argument(
            "--metadata_filename_formatter",
            nargs="?",
            default=None,
            help="""Specify filename format module_name that should be used for
                            metadata output, where filename_module_name is
                            'myfilemodule' where:
                                geoips.filename_formatters.myfilemodule
                            would be the appropriate entry point""",
        )
    if arglist is None or "metadata_filename_formatter_kwargs" in arglist:
        procflow_group.add_argument(
            "--metadata_filename_formatter_kwargs",
            nargs="?",
            default={},
            type=literal_eval,
            help="""Specify filename format kwargs that should be used for
                            this metadata_filename_formatter.
                            Should be formatted as a json dictionary string""",
        )

    if arglist is None or "output_formatter" in arglist:
        procflow_group.add_argument(
            "--output_formatter",
            nargs="?",
            default=None,
            help="""Specify output format module_name that should be used for this file,
                    each output_formatter is 'output_formatters.imagery_annotated' where
                    from geoips*.output_formatters.imagery_annotated import
                    imagery_annotated would be the appropriate import statement""",
        )
    if arglist is None or "output_formatter_kwargs" in arglist:
        procflow_group.add_argument(
            "--output_formatter_kwargs",
            nargs="?",
            default={},
            type=literal_eval,
            help="""Specify output format kwargs that should be used for this
                    output_formatter. should be formatted as a json dictionary string,
                    ie: '{"title_formatter": "tc_copyright", "title_copyright": "NRL"}'
                    """,
        )

    if arglist is None or "metadata_output_formatter" in arglist:
        procflow_group.add_argument(
            "--metadata_output_formatter",
            nargs="?",
            default=None,
            help="""Specify output format module_name that should be used for
                    metadata output, each output_formatter is 'myoutputmodule' where
                        from geoips.output_formatters.myoutputmodule.myoutputmodule
                    would be the appropriate entry point""",
        )
    if arglist is None or "metadata_output_formatter_kwargs" in arglist:
        procflow_group.add_argument(
            "--metadata_output_formatter_kwargs",
            nargs="?",
            default={},
            type=literal_eval,
            help="""Specify output format kwargs that should be used for this metadata.
                    Should be formatted as a json dictionary string.""",
        )

    if arglist is None or "output_config" in arglist:
        procflow_group.add_argument(
            "--output_config",
            nargs="?",
            default=None,
            help="""Specify YAML config file holding output modile names and
                                          their respective filename modules""",
        )

    if arglist is None or "no_presectoring" in arglist:
        procflow_group.add_argument(
            "--no_presectoring",
            action="store_true",
            help="""If true, do not pre-sector data prior to running the algorithm.
                    This is less efficient, but allows the original dataset to
                    be passed to the algorithm in full.""",
        )

    if arglist is None or "output_checker_name" in arglist:
        procflow_group.add_argument(
            "--output_checker_name",
            default=None,
            help="""Output Checker Name Override.
                    - By default, GeoIPS determines the appropriate
                      output_checker plugin to use for comparisons based on
                      the file extension of the output.
                    - If a particular file extension is not explicitly
                      supported, but is expected to be able to utilize one of
                      the existing output checkers (e.g. using the "text"
                      output checker with a CSV file), this argument allows the
                      name of the checker to be passed in to override the default.
                    - I.e., to force the use of the "text" ouput checker with a
                      product that produces a CSV output output_checker plugin,
                      you would pass:
                        --output_checker_name text
                    """,
        )

    if arglist is None or "output_checker_kwargs" in arglist:
        procflow_group.add_argument(
            "--output_checker_kwargs",
            default={},
            type=literal_eval,
            help="""Output Checker Keyword Arguments.
                    - These keyword arguments get passed through directly to the
                      outputs_match method on the output_checker plugin.
                    - This should be formatted as a json dictionary string,
                      with the first level being the plugin name that the
                      keyword arguments apply to, and the second level being
                      a dictionary of keyword argument name/value pairs.
                    - Ie, for the "threshold" keyword argument in the "image"
                      output_checker plugin, you would pass:
                        --output_checker_kwargs '{"image": {"threshold": "0.05"}}'
                    """,
        )

    rdr_group = parser.add_argument_group(title="Data reader specifications")

    if arglist is None or "reader_name" in arglist:
        rdr_group.add_argument(
            "--reader_name",
            default=None,
            help="""If --reader_name is passed, the specific reader will be located in
                    geoips*.readers.myreader_name.myreader_name,
                    The reader_name string should be the reader module name (no .py)""",
        )
    if arglist is None or "reader_kwargs" in arglist:
        sect_group.add_argument(
            "--reader_kwargs",
            nargs="?",
            default=None,
            type=literal_eval,
            help="""Specify reader kwargs that should be used for
                            this reader. Should be formatted as a json
                            dictionary string""",
        )

    if arglist is None or "bg_product_name" in arglist:
        rdr_group.add_argument(
            "--bg_product_name",
            default=None,
            help="""Product to use for background imagery""",
        )
    if arglist is None or "bg_reader_name" in arglist:
        rdr_group.add_argument(
            "--bg_reader_name",
            default=None,
            help="""If --bg_reader_name is passed, the specific reader will be located
                    in geoips*.readers.myreader_name.myreader_name, The bg_reader_name
                    string should be the reader module name (no .py)""",
        )
    if arglist is None or "bg_fnames" in arglist:
        rdr_group.add_argument(
            "--bg_fnames",
            nargs="*",
            default=None,
            help="""Specify filenames to use for background imagery.
                    If --bg_reader_name included, use specific reader for
                    reading background datafiles.""",
        )

    plt_group = parser.add_argument_group(title="Plotting parameter specifications")
    if arglist is None or "gridline_annotator" in arglist:
        plt_group.add_argument(
            "--gridline_annotator",
            default=None,
            help="""If --gridline_annotator is passed, the specific gridline
                    params will be located in
                    geoips*.image_utils.plotting_params.gridlines.gridline_annotator,
                    The gridline_annotator string should be the base gridline name
                    (no .yaml)""",
        )
    if arglist is None or "feature_annotator" in arglist:
        plt_group.add_argument(
            "--feature_annotator",
            default=None,
            help="""If --feature_annotator is passed, the specific feature
                    annotations will be located in
                    geoips*.plugins.yaml.feature_annotators.<feature_annotator>,
                    The feature_annotator string should be the base feature annotator
                    name (no .yaml)""",
        )

    if arglist is None or "model_reader_name" in arglist:
        rdr_group.add_argument(
            "--model_reader_name",
            default=None,
            help="""If --model_reader_name is passed, the specific reader
                    will be located in
                    geoips*.readers.my_reader_name.my_reader_name,
                    The model_reader_name string should be the reader module name
                    (no .py)""",
        )
    if arglist is None or "model_fnames" in arglist:
        rdr_group.add_argument(
            "--model_fnames",
            nargs="*",
            default=None,
            help="""Specify filenames to use for NWP model data.
                    If --model_reader_name included, use specific reader for
                    reading model datafiles.""",
        )
    if arglist is None or "buoy_reader_name" in arglist:
        rdr_group.add_argument(
            "--buoy_reader_name",
            default=None,
            help="""If --buoy_reader_name is passed, the specific reader
                    will be located in
                    geoips*.readers.my_reader_name.my_reader_name,
                    The buoyreadername string should be the reader module name
                    (no .py)""",
        )
    if arglist is None or "buoy_fnames" in arglist:
        rdr_group.add_argument(
            "--buoy_fnames",
            nargs="*",
            default=None,
            help="""Specify filenames to use for buoy data.
                    If --buoy_reader_name included, use specific reader for
                    reading buoy datafiles.""",
        )
    if arglist is None or "aeorsol_reader_name" in arglist:
        rdr_group.add_argument(
            "--aerosol_reader_name",
            default=None,
            help="""If --model_reader_name is passed, the specific reader
                    will be located in
                    geoips*.readers.my_reader_name.my_reader_name,
                    The model_reader_name string should be the reader module name
                    (no .py)""",
        )
    if arglist is None or "aerosol_fnames" in arglist:
        rdr_group.add_argument(
            "--aerosol_fnames",
            nargs="*",
            default=None,
            help="""If --aerosol_reader_name included, use specific reader for
                    reading aerosol model datafiles.""",
        )

    fusion_group = parser.add_argument_group(
        title="Options for specifying fusion products"
    )
    if arglist is None or "fuse_files" in arglist:
        fusion_group.add_argument(
            "--fuse_files",
            action="append",
            nargs="+",
            default=None,
            help="""Provide additional files required for fusion product.
                    Files passed under this flag MUST be from the same source.
                    fuse_files may be passed multiple times.  Reader name for
                    these files is specified with the fuse_reader flag.""",
        )
        fusion_group.add_argument(
            "--fuse_reader",
            action="append",
            default=None,
            help="""Provide the reader name for files passed under the
                    fuse_files flag. Only provide one reader to this flag.
                    If multiple fuse_files flags are passed, the same number of
                    fuse_readers must be passed in the same order.""",
        )
        fusion_group.add_argument(
            "--fuse_reader_kwargs",
            action="append",
            default=None,
            type=literal_eval,
            help="""Provide the reader kwargs for files passed under the
                    fuse_files flag. Should be formatted as a json dictionary string.
                    Only provide one json dict str to this flag.
                    If multiple fuse_files flags are passed, the same number of
                    fuse_reader_kwargs must be passed in the same order.""",
        )
        fusion_group.add_argument(
            "--fuse_product",
            action="append",
            default=None,
            help="""Provide the product name for files passed under the
                    fuse_files flag. Only provide one product to this flag.
                    If multiple fuse_files flags are passed, the same number of
                    fuse_products must be passed in the same order.""",
        )
        fusion_group.add_argument(
            "--fuse_resampled_read",
            action="append",
            default=None,
            help="""Identify whether the reader specified for
                    --fuse_files / --fuse_reader perform a resampled_read.
                    If not specified, a resampled read will NOT be performed.
                    If multiple fuse_files flags are passed, either the same
                    number of fuse_resampled_read flags must be passed in the
                    same order, or no fuse_resampled_read flags can be passed.""",
        )
        fusion_group.add_argument(
            "--fuse_sectored_read",
            action="append",
            default=None,
            help="""Identify whether the reader specified for --fuse_files/--fuse_reader
                    perform a sectored_read.  If not specified, a sectored_read will NOT
                    be performed.  If multiple fuse_files flags are passed, either
                    the same number of fuse_sectored_read flags must be passed
                    in the same order, or no fuse_sectored_read flags can be passed.""",
        )
        fusion_group.add_argument(
            "--fuse_self_register_dataset",
            action="append",
            default=None,
            help="""Identify the DATASET of the data to which the associated
                    fuse_files should be registerd. If not specified, data will
                    not be registered to a dataset, but the requested area_def.
                    If multiple fuse_files flags are passed, either the same
                    number of fuse_self_register_dataset flags must be passed in
                    the same order, or no fuse_self_register_dataset flags can
                    be passed.""",
        )
        fusion_group.add_argument(
            "--fuse_self_register_source",
            action="append",
            default=None,
            help="""Identify the SOURCE of the data to which the associated
                    fuse_files should be registerd If not specified, data will
                    not be registered to a dataset, but the requested. area_def.
                    If multiple fuse_files flags are passed, either the same
                    number of fuse_self_register_source flags must be passed in
                    the same order, or no fuse_self_register_source flags can be
                    passed.""",
        )
        fusion_group.add_argument(
            "--fuse_self_register_platform",
            action="append",
            default=None,
            help="""Identify the SOURCE of the data to which the associated
                    fuse_files should be registerd. If not specified, data will
                    not be registered to a dataset, but the requested area_def.
                    If multiple fuse_files flags are passed, either the same
                    number of fuse_self_register_source flags must be passed in
                    the same order, or no fuse_self_register_source flags can be
                    passed.""",
        )

    prod_db_group = parser.add_argument_group(title="Product database specifications")
    if arglist is None or "product_db" in arglist:
        prod_db_group.add_argument(
            "--product_db",
            action="store_true",
            help="Use product database to store product paths created by the procflow",
        )
        prod_db_group.add_argument(
            "--product_db_writer",
            default=None,
            help="""If --product_db_writer is passed, the specific product
                    database writer will be located in
                    geoips*.plugins.modules.postgres_database.
                        mywriter_name.mywriter_name,
                    The writer_name string should be the reader module name
                    (no .py)""",
        )
        prod_db_group.add_argument(
            "--product_db_writer_kwargs",
            default=None,
            type=literal_eval,
            help="""Provide the product db writer kwargs for the plugin passed under the
                    product_db_writer flag. Should be formatted as a json dictionary
                    string. Only provide one json dict str to this flag.""",
        )
        prod_db_group.add_argument(
            "--product_db_writer_override",
            nargs="?",
            default={},
            type=literal_eval,
            help="""Specify product database writer that should be used for each
                    available sector should be formatted as a json dictionary
                    string.""",
        )
        prod_db_group.add_argument(
            "--store_checkpoint_statistics",
            action="store_true",
            help="""Store the tracked resource usage statistics for each checkpoint to
                    the backend database.""",
        )

    composite_group = parser.add_argument_group(title="Image composite kwargs")
    if arglist is None or "composite_output_kwargs_override" in arglist:
        composite_group.add_argument(
            "--composite_output_kwargs_override",
            nargs="?",
            default={},
            type=literal_eval,
            help="""Specify product composite kwargs that should be used for each
                    available sector output. Should be formatted as a json dictionary
                    string.""",
        )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read EWS-G data.

This EWS-G(Electro-Optical Infrared Weather System - Geostationary) reader is
designed for reading theEWS-G data files (EWS-G is renamed from GOES-13).
The reader is only using the python functions and xarray variables.
The reader is based on EWS-G data in netcdf4 format.

 V1.0:  NRL-Monterey, 02/25/2021

EWS-G file information::

    Example of the gvar filename:   2020.1212.0012.goes-13.gvar.nc

    Note that channel-3 is not available for EWS-G.
      gvar_Ch3(TIR=5.8-7.3um, ctr=6.48um,4km): unit=temp-deg(C), scale_factor=0.01

    variables:
      gvar_Ch1(VIS=0.55-0.75um, ctr=0.65um,1km): unit=albedo*100,  scale_factor=0.01
      gvar_Ch2(MWIR=3.8-4.0um,  ctr=3.9um, 4km): unit=temp-deg(C), scale_factor=0.01
      gvar_Ch4(TIR=10.2-11.2um, ctr=10.7um,4km): unit=temp-deg(C), scale_factor=0.01
      gvar_Ch6(TIR=12.9-13.7um, ctr=13.3um 4km): unit=temp-deg(C), scale_factor=0.01
      latitude: unit=degree
      longitude:unit=degree
      sat_zenith: unit=degree
      sun_zenith: unit=degree
      rel_azimuth:unit=degree

      variable array definition:  var(scan,pix); scan-->lines, pix-->samples

    attributes: many
"""
# Python Standard Libraries
import logging
import os

# Third-Party Libraries
import calendar
import numpy as np
import pandas as pd
import xarray as xr

from geoips.interfaces import readers
from geoips.filenames.base_paths import PATHS as gpaths
from geoips.utils.context_managers import import_optional_dependencies

# If this reader is not installed on the system, don't fail altogether, just skip this
# import. This reader will not work if the import fails, and the package will have to be
# installed to process data of this type.

LOG = logging.getLogger(__name__)

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import netCDF4 as ncdf


# @staticmethod                                     # not sure where it is was used?

# gvar_ch6 has only half scanlines of other channels (1,2,4), we temporally do not read
# the ch6 in. We will modify this reader if gvar_ch6 is needed in the future.
VARLIST = [
    "gvar_ch1",
    "gvar_ch2",
    "gvar_ch4",
    "latitude",
    "longitude",
    "sun_zenith",
    "sat_zenith",
    "rel_azimuth",
]

# setup needed to convert var_name used in geoips: i.e., solar_zenith_angle (not
# sun_zenith) is used.
xvarnames = {
    "sun_zenith": "solar_zenith_angle",
    "sat_zenith": "satellite_zenith_angle",
    "rel_azimuth": "satellite_azimuth_angle",
}

interface = "readers"
family = "standard"
name = "ewsg_netcdf"
source_names = ["gvar"]


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    roi=None,
    geolocation_cache_backend=gpaths["GEOIPS_GEOLOCATION_CACHE_BACKEND"],
    cache_chunk_size=None,
    cache_solar_angles=False,
    geolocation_only=False,
    resource_tracker=None,
):
    """Read EWS-G data in netcdf4 format.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.
    roi: radius of influence (unit in meter), used in interpolation scheme
        * Default=None, i.e., if not defined in the yaml file
        * Defined in yaml file where variables and tuning parameters are set

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    return readers.read_data_to_xarray_dict(
        fnames,
        call_single_time,
        metadata_only,
        chans,
        area_def,
        self_register,
        roi=roi,
        geolocation_cache_backend=geolocation_cache_backend,
        cache_chunk_size=cache_chunk_size,
        resource_tracker=resource_tracker,
        cache_solar_angles=cache_solar_angles,
        geolocation_only=geolocation_only,
    )


def call_single_time(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    roi=None,
    geolocation_cache_backend="memmap",
    cache_chunk_size=None,
    cache_solar_angles=False,
    geolocation_only=False,
    resource_tracker=None,
):
    """Read EWS-G data in netcdf4 format for one or more files.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.
    roi: radius of influence (unit in meter), used in interpolation scheme
        * Passed in from Call function
        * Default=None, i.e., if not defined in the yaml file
        * Defined in yaml file where variables and tuning parameters are set
    geolocation_cache_backend : str
        * Specify to use either memmap or zarray to store pre-calculated geolocation
          data.
        * Not currently used, since these files have all required geolocation data in
          the netCDF file.
    cache_chunk_size : int
        * Specify chunck size if using zarray to store pre-calculated geolocation data.
        * Not currently used, since geolocation_cache_backend is currently not used.
    resource_tracker: geoips.utils.memusg.PidLog object
        * Track resource usage using the PidLog class object from geoips.utils.memusg.
        * The PidLog.track_resource_usage method allows us to snapshot the memory usage
          for the PID associated with the geoips call. The time and stats of the
          snapshot are recorded, and can be accessed using the
          PidLog.checkpoint_usage_stats method.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # --------------- loop input files ---------------
    xarray_ewsg = xr.Dataset()
    xarray_ewsg.attrs["source_file_names"] = []
    # LOG.info('Requested Channels: %s', chans)

    for fname in fnames:
        # check for a correct goes-13 data file
        data_name = os.path.basename(fname).split("_")[-1].split(".")[-1]
        if data_name != "nc":
            LOG.info("Warning: EWS-G data type:  data_type=", data_name)
            raise

        # open the paired input files
        ncdf_file = ncdf.Dataset(str(fname), "r")
        LOG.info("    Trying file %s", fname)

        if ncdf_file.satellite in ["goes-13", "goes-14", "goes-15"]:
            LOG.info("found a NOAA EWS-G data file")
        else:
            LOG.info("not a NOAA EWS-G data file: skip it")
            raise ValueError(
                f"Found {ncdf_file.satellite}, "
                "expected either goes-13, goes-14, or goes-15"
            )

        # setup attributes
        # use fname to get an initial info of  year, month, day
        # test hour/minute/second of "start_datetime" info from start_time (second of
        # the day). add scan_time info to determine the "end_datetime". scan_time has
        # time of scans for this input data file(one obs).  If end_time of this data >24
        # hr, modify value of "day" from fname.  Note: scan_time units are seconds from
        # (start_time + time_adjust)

        # date info from fname
        #        1  2  3  4  5  6  7  8  9  10 11 12
        days_mo = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        days_mo_lp = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]  # Leap year

        # date information is not contained in the data so we have to get it from
        # filename
        data_name = os.path.basename(fname)
        yr = int(data_name.split(".")[0])
        mo = int(data_name.split(".")[1][0:2])
        dy = int(data_name.split(".")[1][2:4])
        # hr = int(data_name.split(".")[2][0:2])
        # mm = int(data_name.split(".")[2][2:4])

        # determine a Leap Year?
        if calendar.isleap(yr):
            days = days_mo_lp[mo - 1]
        else:
            days = days_mo[mo - 1]

        # second of the date for this file
        start_time = (
            ncdf_file.start_time + ncdf_file.time_adjust + ncdf_file.scan_time[0]
        )
        end_time = (
            ncdf_file.start_time + ncdf_file.time_adjust + ncdf_file.scan_time[29]
        )

        yr_s = yr
        yr_e = yr
        mo_s = mo
        mo_e = mo
        dy_s = dy
        dy_e = dy
        hr_s = int(start_time / 3600)
        mm_s = int((start_time - hr_s * 3600) / 60)
        ss_s = int(start_time - (hr_s * 3600 + mm_s * 60))
        hr_e = int(end_time / 3600)
        mm_e = int((end_time - hr_e * 3600) / 60)
        ss_e = int(end_time - (hr_e * 3600 + mm_e * 60))

        if hr_s >= 24:
            # dy_s_ = dy + 1  # forward to the next date
            hr_s = hr_s - 24
            if dy_s > days:  # move to next mon
                mo_s = mo + 1
                if mo_s > 12:  # move to near year
                    yr_s = yr + 1
        if hr_e >= 24:
            # dy_e_ = dy + 1  # forward to the next date
            hr_e = hr_e - 24
            if dy_e > days:  # move to next mon
                mo_e = mo + 1
                if mo_e > 12:  # move to near year
                    yr_e = yr + 1

        start_scan = "%04d%02d%02d%02d%02d%02d" % (yr_s, mo_s, dy_s, hr_s, mm_s, ss_s)
        end_scan = "%04d%02d%02d%02d%02d%02d" % (yr_e, mo_e, dy_e, hr_e, mm_e, ss_e)

        # convert date in required format
        Start_date = pd.to_datetime(start_scan, format="%Y%m%d%H%M%S").to_pydatetime()
        End_date = pd.to_datetime(end_scan, format="%Y%m%d%H%M%S").to_pydatetime()

        xarray_ewsg.attrs["start_datetime"] = Start_date
        xarray_ewsg.attrs["end_datetime"] = End_date
        # Goes VARiable (gvar) data is on GOES-12/13 - older GOES satellites were gvissr
        # ncdf_file.sensor_name is gvissr - I think that is a mistake on
        # Terascan's part - hold out from years ago
        xarray_ewsg.attrs["source_name"] = "gvar"

        if ncdf_file.satellite == "goes-13":
            xarray_ewsg.attrs["platform_name"] = "ews-g1"
        if ncdf_file.satellite == "goes-15":
            xarray_ewsg.attrs["platform_name"] = "ews-g2"
        if ncdf_file.satellite == "goes-14":
            xarray_ewsg.attrs["platform_name"] = "ews-g3"
        xarray_ewsg.attrs["legacy_platform_name"] = ncdf_file.satellite
        xarray_ewsg.attrs["data_provider"] = "noaa"
        xarray_ewsg.attrs["source_file_names"] = [
            os.path.basename(fname) for fname in fnames
        ]

        # MTIFs need to be "prettier" for PMW products, so 2km resolution for
        # final image
        xarray_ewsg.attrs["sample_distance_km"] = 2
        if roi is not None:
            xarray_ewsg.attrs["interpolation_radius_of_influence"] = roi
        else:
            xarray_ewsg.attrs["interpolation_radius_of_influence"] = 3000
        # Just return the metadata
        if metadata_only:
            # close the files
            ncdf_file.close()
            return {"METADATA": xarray_ewsg}
        # Otherwise, get the actual data, as well as the metadata
        # *************** input VIRRS variables  and output xarray required by geo

        # for varname in ncdf_file.variables.keys():
        for var in VARLIST:
            if resource_tracker:
                key = f"READ CHAN: abi_netcdf_chan_{var}"
                if area_def:
                    key += f"_{area_def.area_id}"
                resource_tracker.track_resource_usage(
                    logstr="MEMUSG",
                    verbose=False,
                    key=key,
                    show_log=False,
                )
            varname = var
            data = ncdf_file[varname]
            # masked_data=np.ma.masked_equal(
            #   ncdf_file[varname],ncdf_file[varname].missing_value
            # )
            masked_data = np.ma.masked_equal(data, data.missing_value)

            if var in xvarnames:
                varname = xvarnames[var]  # rename zenith/azimuth-related variables

            xarray_ewsg[varname] = xr.DataArray(masked_data)
            """
            # scale_factor should not be applied
            if 'scale_factor' in data.ncattrs():
                # apply scale_factor correction
                #xarray_ewsg[varname]=xarray_ewsg[varname]*ncdf_file[varname].scale_factor  # NOQA
                xarray_ewsg[varname]=xarray_ewsg[varname]*data.scale_factor
            """
            # convert unit from degree to Kelvin for ch2, ch4, and ch6 (ch1 is in unit
            # of albedo)
            if varname in ["gvar_ch2", "gvar_ch4", "gvar_ch6"]:
                xarray_ewsg[varname] = xarray_ewsg[varname] + 273.15
                xarray_ewsg[varname].attrs["units"] = "Kelvin"
            if resource_tracker:
                resource_tracker.track_resource_usage(
                    logstr="MEMUSG", verbose=False, key=key
                )

        # close the files
        ncdf_file.close()

    return {"LOW": xarray_ewsg, "METADATA": xarray_ewsg[[]]}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module for handling cached files in GeoIPS.

This modules provides functions to manage cache files in GeoIPS.
Cache files will be stored in the user cache directory, which is platform-dependent. The
correct cache directory is determined using the GEOIPS_CACHE_DIR environment variable
which defaults to `platformdirs.user_cache_dir("geoips")` if not set.
"""

import os
import yaml
import json
from logging import getLogger
from geoips.filenames.base_paths import PATHS

LOG = getLogger(__name__)


def source_modified(source, dest):
    """Check if the source file was modified more recently than the destination file.

    This uses os.path.getmtime() to determine whether the source file has been modified
    more recently than the destination file. Returns True if the source file was
    modified more recently than the destination file, False otherwise.

    Parameters
    ----------
    source: str
        The path to the source file we are monitoring for changes.
    dest: str
        The path to the destination file we will update if the source file changes.

    Returns
    -------
    bool
        True if the source file was modified more recently than the destination file,
        False otherwise.
    """
    if not os.path.exists(dest) or os.path.getmtime(source) > os.path.getmtime(dest):
        return True
    return False


def create_cached_json_from_yaml(source, cache_dir=None):
    """Create a cached JSON file from a YAML file.

    This function reads a YAML file and writes its contents to a JSON file in the user
    cache directory. The JSON file will be created if it does not already exist, or
    updated if it does.

    Parameters
    ----------
    source: str
        The path to the source YAML file.
    cache_dir: str, optional
        The path to the cache directory. If not provided, the default user cache
        directory will be used.

    Returns
    -------
    str
        The path to the cached JSON file.
    """
    if not cache_dir:
        cache_dir = PATHS["GEOIPS_CACHE_DIR"]

    os.makedirs(cache_dir, exist_ok=True)
    dest = os.path.join(cache_dir, os.path.basename(source).replace(".yaml", ".json"))

    if source_modified(source, dest):
        with open(source, "r") as yaml_file:
            data = yaml.safe_load(yaml_file)

        with open(dest, "w") as json_file:
            json.dump(data, json_file, indent=4)
    return dest


def get_cached_json(source, cache_dir=None):
    """Get the cached JSON file corresponding to a YAML file.

    Some files in GeoIPS are stored in YAML format, but we want to use JSON at runtime
    because loading JSON is faster than loading YAML. This function checks if the cached
    JSON file exists and is up to date. If it does not exist or is out of date, it
    creates a new cached JSON file from the YAML file. It then returns the contents of
    the cached JSON file.

    Parameters
    ----------
    source: str
        The path to the source YAML file.
    cache_dir: str, optional
        The path to the cache directory. If not provided, the default user cache
        directory will be used.
    """
    cache_file = create_cached_json_from_yaml(source)
    with open(cache_file, "r") as json_file:
        return json.load(json_file)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Full NetCDF output format, writing out all channels in full dataset."""
import logging

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "xarray_data"
name = "netcdf_full"


def call(
    xarray_obj,
    product_names,
    output_fnames,
    clobber=False,
    use_compression=True,
    compression_kwargs=None,
):
    """Write GeoIPS style NetCDF to disk."""
    from geoips.plugins.modules.output_formatters.netcdf_xarray import (
        write_xarray_netcdf,
    )

    for ncdf_fname in output_fnames:
        write_xarray_netcdf(
            xarray_obj,
            ncdf_fname,
            clobber=clobber,
            use_compression=use_compression,
            compression_kwargs=compression_kwargs,
        )
    return output_fnames



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for retrieving commandline instructions for the GeoIPS CLI."""

import os
from os.path import dirname, exists

import pytest
import tempfile

from geoips.utils.cache_files import get_cached_json
from geoips.commandline.commandline_interface import GeoipsCLI


# This points to the instructions directory for the unit tests. These are test
# instructions that are not used in the actual codebase.
INSTRUCT_DIR = f"{str(dirname(__file__))}/cmd_instructions"


def generate_id(dir_name):
    """Generate a test id for the instructions directory being tested."""
    return dir_name


@pytest.mark.parametrize("dir_name", os.listdir(INSTRUCT_DIR), ids=generate_id)
def test_instruction_cases(dir_name):
    """Check that the instructions found in dir_name pass or raise the correct error.

    Parameters
    ----------
    dir_name: str
        - The directory name containing the cmd_instructions we want to test
        - Will be testing:
            - missing both .yaml and .json instructions
            - missing .json instructions
            - missing .yaml instructions
            - invalid .json instructions
            - invalid .yaml instructions
            - newer .yaml file than the .json file
            - newer .json file than the .yaml file
    """
    # This will take the place of the user cache directory for the unit tests.
    # This is where the cached JSON files will be stored.
    # This is a temporary directory that will be deleted when the test is complete.
    cache_dir = tempfile.mkdtemp()

    # Path to the directory containing the YAML instructions
    cmd_dir = f"{INSTRUCT_DIR}/{dir_name}"
    cmd_yaml = f"{cmd_dir}/cmd_instructions.yaml"
    cmd_json = f"{cache_dir}/cmd_instructions.json"

    if "invalid" in dir_name:
        # Load the invalid instructions
        cmd_instructions = get_cached_json(
            f"{cmd_dir}/cmd_instructions.yaml",
            cache_dir=cache_dir,
        )
        # Ensure that a key error is raised
        GeoipsCLI(cmd_instructions=cmd_instructions)
    elif "missing" in dir_name:
        if dir_name == "json_missing":
            cmd_instructions = get_cached_json(cmd_yaml, cache_dir=cache_dir)
            assert exists(cmd_json)
        else:
            with pytest.raises(FileNotFoundError):
                get_cached_json(cmd_yaml, cache_dir=cache_dir)

    # Clean up
    try:
        os.remove(cmd_json)
    except FileNotFoundError:
        pass
    os.rmdir(cache_dir)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard GeoIPS xarray dictionary based FCI NetCDF data reader."""

from datetime import datetime, timedelta
import logging
import numpy as np
import os
from pathlib import Path
from pyorbital.astronomy import sun_earth_distance_correction
import xarray
import zarr

from geoips.interfaces import readers
from geoips.filenames.base_paths import PATHS as gpaths
from geoips.plugins.modules.readers.utils.geostationary_geolocation import (
    check_geolocation_cache_backend,
    get_geolocation_cache_filename,
    get_geolocation,
    AutoGenError,
)
from geoips.utils.context_managers import import_optional_dependencies

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import numexpr as ne
    from satpy.scene import Scene
    import netCDF4 as ncdf

    # Need to uncomment if HDF5_PLUGIN_PATH env variable is not set upstream,
    # though you will still not be able to interact with the data after using Scene.load
    # Until this works out of the box with out the need to set HDF5_PLUGIN_PATH
    # import hdf5plugin

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "fci_netcdf"
source_names = ["fci"]

nprocs = 6

try:
    ne.set_num_threads(nprocs)
except Exception:
    LOG.info(
        f"Failed numexpr.set_num_threads in {__file__}. "
        "If numexpr is not installed and you need it, install it."
    )

DONT_AUTOGEN_GEOLOCATION = False
if os.getenv("DONT_AUTOGEN_GEOLOCATION"):
    DONT_AUTOGEN_GEOLOCATION = True

# Needed until the satpy devs can fix this issue:
# https://github.com/pytroll/satpy/issues/3067
# Used to convert from wavenumber (mW·m⁻²·sr⁻¹·(cm⁻¹)⁻¹)
# to wavelength (W·m⁻²·sr⁻¹·µm⁻¹)
RADIANCE_CONVERSION_COEFFICIENTS = {
    "B01Rad": 50.26570892,
    "B02Rad": 38.74919891,
    "B03Rad": 24.66601944,
    "B04Rad": 13.51755047,
    "B05Rad": 11.98462009,
    "B06Rad": 5.261388779,
    "B07Rad": 3.84947896,
    "B08Rad": 1.965775013,
    "B09Rad": 0.697183013,
    "B10Rad": 0.2536683083,
    "B11Rad": 0.1817249954,
    "B12Rad": 0.1306722015,
    "B13Rad": 0.1071347967,
    "B14Rad": 0.0899727419,
    "B15Rad": 0.06602230668,
    "B16Rad": 0.05665054172,
    "HRB03Rad": 24.66601944,
    "HRB09Rad": 0.697183013,
    "HRB14Rad": 0.0899727419,
}

DATASET_INFO = {
    # HIGH resolution = 0.5km at nadir
    "HIGH": ["vis_06_hr", "nir_22_hr"],
    # MED resolution = 1.0km at nadir
    "MED": [
        "vis_04",
        "vis_05",
        "vis_06",
        "vis_08",
        "vis_09",
        "nir_13",
        "nir_16",
        "nir_22",
        "ir_38_hr",
        "ir_105_hr",
    ],
    # LOW resolution = 2.0km at nadir
    "LOW": ["ir_38", "wv_63", "wv_73", "ir_87", "ir_97", "ir_105", "ir_123", "ir_133"],
}

BAND_MAP = {
    "B01Ref": "vis_04",
    "B01Rad": "vis_04",
    "B02Ref": "vis_05",
    "B02Rad": "vis_05",
    "B03Ref": "vis_06",
    "B03Rad": "vis_06",
    "B04Ref": "vis_08",
    "B04Rad": "vis_08",
    "B05Ref": "vis_09",
    "B05Rad": "vis_09",
    "B06Ref": "nir_13",
    "B06Rad": "nir_13",
    "B07Ref": "nir_16",
    "B07Rad": "nir_16",
    "B08Ref": "nir_22",
    "B08Rad": "nir_22",
    "B09BT": "ir_38",
    "B09Rad": "ir_38",
    "B10BT": "wv_63",
    "B10Rad": "wv_63",
    "B11BT": "wv_73",
    "B11Rad": "wv_73",
    "B12BT": "ir_87",
    "B12Rad": "ir_87",
    "B13BT": "ir_97",
    "B13Rad": "ir_97",
    "B14BT": "ir_105",
    "B14Rad": "ir_105",
    "B15BT": "ir_123",
    "B15Rad": "ir_123",
    "B16BT": "ir_133",
    "B16Rad": "ir_133",
    "HRB03Ref": "vis_06_hr",
    "HRB03Rad": "vis_06_hr",
    "HRB09BT": "ir_38_hr",
    "HRB09Rad": "ir_38_hr",
    "HRB14BT": "ir_105_hr",
    "HRB14Rad": "ir_105_hr",
}

CHAN_MAP = {val: key for key, val in BAND_MAP.items()}

ALL_GEO_VARS = [
    "solar_zenith_angle",
    "satellite_zenith_angle",
    "solar_azimuth_angle",
    "satellite_azimuth_angle",
    "latitude",
    "longitude",
]

BADVALS = {
    "Off_Of_Disk": -999.9,
    "Conditional": -999.8,
    "Out_Of_Valid_Range": -999.7,
    "No_Value": -999.6,
    "Unitialized": -9999.9,
}


def get_geolocation_metadata(xarray_attrs):
    """Get geolocation metadata."""
    geomet = {}
    geomet["platform_name"] = xarray_attrs["platform_name"]
    geomet["source_name"] = xarray_attrs["source_name"]  # sensor == fci
    # Used for cached filenames
    if "orbital_parameters" in xarray_attrs:
        # xarray from converted satpy.scene
        geomet["lon0"] = xarray_attrs["orbital_parameters"][
            "satellite_nominal_longitude"
        ]
        geomet["num_lines"] = xarray_attrs["area"].y_size
        geomet["num_samples"] = xarray_attrs["area"].x_size
        geomet["scan_mode"] = xarray_attrs["area"].area_id
    else:
        geomet["lon0"] = xarray_attrs["mtg_geos_projection"][
            "longitude_of_projection_origin"
        ]
        geomet["Re_m"] = xarray_attrs["mtg_geos_projection"]["semi_major_axis"]
        geomet["Rp_m"] = xarray_attrs["mtg_geos_projection"]["semi_minor_axis"]
        geomet["scan_mode"] = xarray_attrs["coverage"]
        geomet["num_lines"] = xarray_attrs["num_lines"]
        geomet["num_samples"] = xarray_attrs["num_samples"]
        geomet["mtg_geos_projection"] = xarray_attrs["mtg_geos_projection"]
    geomet["scene"] = geomet["scan_mode"]
    geomet["start_datetime"] = xarray_attrs["start_datetime"]
    geomet["end_datetime"] = xarray_attrs["end_datetime"]
    if "H_m" not in geomet:
        geomet["H_m"] = 42164000.0
        geomet["Re_m"] = 6378137.0
        geomet["Rp_m"] = 6356752.298215968
    geomet["H_km"] = geomet["H_m"] / 1000.0
    geomet["Re_km"] = geomet["Re_m"] / 1000.0
    geomet["Rp_km"] = geomet["Rp_m"] / 1000.0
    geomet["roi_factor"] = 5
    return geomet


def get_latitude_longitude(
    metadata,
    x_1d,
    y_1d,
    BADVALS,
    sect=None,
    geolocation_cache_backend="zarr",
    chunk_size=None,
    resource_tracker=None,
):
    """
    Get latitudes and longitudes.

    This routine accepts a dictionary containing metadata as read from a NCDF4
    format file, and returns latitudes and longitudes for a full disk.
    """
    check_geolocation_cache_backend(geolocation_cache_backend)
    # If the filename format needs to change for the pre-generated geolocation
    # files, please discuss prior to changing.  It will force recreation of all
    # files, which can be problematic for large numbers of sectors
    fname = get_geolocation_cache_filename(
        "GEOLL",
        metadata,
        geolocation_cache_backend=geolocation_cache_backend,
        chunk_size=chunk_size,
    )
    if resource_tracker is not None:
        key = Path(fname).name
        if sect:
            key += f"_{sect.area_id}"
        resource_tracker.track_resource_usage(logstr="MEMUSG", verbose=False, key=key)
    if not Path(fname).exists():
        if sect is not None and DONT_AUTOGEN_GEOLOCATION and "tc2019" not in sect.name:
            msg = (
                f"GETGEO Requested NO AUTOGEN GEOLOCATION. "
                f"Could not create latlonfile for ad {metadata['scene']}: {fname}"
            )
            LOG.error(msg)
            raise AutoGenError(msg)
        # Anywhere you see NOQA or noqa: F841, this is added since the variables are
        # used however they are not recognized by flake8 linter under numexpr.evaluate()
        # Constants
        pi = np.pi
        # Must include rad2deg variable, because it is used within the
        # numexpr command below.  flake8 does not recognize it as being
        # used, so must include # NOQA flag
        rad2deg = 180.0 / pi  # NOQA
        Re = metadata["Re_km"]  # Earth equatorial radius (km)
        Rp = metadata["Rp_km"]  # Earth polar radius (km)
        # The parameter h in the equations above refers to the geostationary radius.
        # The geostationary radius is the distance from the Earth’s centre to the
        # satellite in geostationary orbit and can be calculated from the sum of the
        # geostationary altitude (35786.4km) and the equatorial Earth radius.
        h = metadata["H_km"]

        s4 = Re**2 / Rp**2  # noqa: F841
        s5 = h**2 - Re**2  # noqa: F841

        lon0 = metadata["lon0"]  # noqa: F841

        x, y = np.meshgrid(x_1d, y_1d)

        cos_x = np.cos(x)
        sin_x = np.sin(x)  # noqa: F841
        cos_y = np.cos(y)
        sin_y = np.sin(y)  # noqa: F841

        sd = ne.evaluate("(h * cos_x * cos_y)**2 - (cos_y**2 + s4 * sin_y**2) * s5")
        bad_mask = sd < 0.0
        sd[bad_mask] = 0.0
        sd **= 0.5

        # Doing inplace operations when variables are no longer needed

        # sd no longer needed
        sn = sd
        ne.evaluate("(h * cos_x * cos_y - sd) / (cos_y**2 +s4 * sin_y**2)", out=sn)

        # cos_x no longer needed
        s1 = cos_x
        ne.evaluate("h - (sn * cos_x * cos_y)", out=s1)

        # Nothing unneed
        s2 = ne.evaluate("sn * sin_x * cos_y")  # noqa: F841

        # sin_y no longer needed
        s3 = cos_y
        ne.evaluate("-sn * sin_y", out=s3)

        # sn no longer needed
        sxy = sn
        ne.evaluate("sqrt(s1**2 + s2**2)", out=sxy)

        # s3 no longer needed
        lats = s3
        ne.evaluate("rad2deg * arctan(s4 * s3 / sxy)", out=lats)

        # s1 no longer needed
        lons = s1
        ne.evaluate("rad2deg * arctan(s2 / s1) + lon0", out=lons)
        lons[lons > 180.0] -= 360

        # Set bad values
        lats[bad_mask] = BADVALS["Off_Of_Disk"]
        lons[bad_mask] = BADVALS["Off_Of_Disk"]
        LOG.info("Done calculating latitudes and longitudes")

        if geolocation_cache_backend == "memmap":
            with open(fname, "w") as df:
                lats.tofile(df)
                lons.tofile(df)
        elif geolocation_cache_backend == "zarr":
            if chunk_size:
                chunks = (chunk_size, chunk_size)
            else:
                chunks = None
            LOG.info("Storing to %s (chunks=%s)", fname, chunks)
            # zarr does NOT have a close method, so you can NOT use the with context.
            zf = zarr.open(fname, mode="w")
            # Assume both arrays have the same shape and dtype
            kwargs = {
                "shape": lats.shape,
                "dtype": lats.dtype,
            }
            # As of Python 3.11, can't pass chunks=None into create_dataset
            if chunks:
                kwargs["chunks"] = chunks
            zf.create_dataset("lats", **kwargs)
            zf.create_dataset("lons", **kwargs)
            zf["lats"][:] = lats
            zf["lons"][:] = lons
    else:
        # Create memmap to the lat/lon file
        # Nothing will be read until explicitly requested
        # We are mapping this here so that the lats and lons are available when
        # calculating satlelite angles
        if geolocation_cache_backend == "memmap":
            shape = (metadata["num_lines"], metadata["num_samples"])
            offset = 8 * metadata["num_samples"] * metadata["num_lines"]
            lats = np.memmap(fname, mode="r", dtype=np.float64, offset=0, shape=shape)
            lons = np.memmap(
                fname, mode="r", dtype=np.float64, offset=offset, shape=shape
            )
        elif geolocation_cache_backend == "zarr":
            LOG.info(
                "GETGEO zarr to {} : lat/lon file for {}".format(
                    fname, metadata["scene"]
                )
            )
            # NOTE zarr does NOT have a close method, so you can NOT use with context.
            zf = zarr.open(fname, mode="r")
            lats = zf["lats"]
            lons = zf["lons"]
    if resource_tracker is not None:
        resource_tracker.track_resource_usage(logstr="MEMUSG", verbose=False, key=key)

    return np.flipud(lats), np.fliplr(lons)


def get_standard_geoips_attrs(file_attrs, area_def):
    """Set the standard geoips attrs.

    Parameters
    ----------
    file_attrs : dict
        Dictionary holding attributes found in source file

    Returns
    -------
    dict
        Dictionary of all source file attrs, along with required GeoIPS attrs
    """
    # Native metadata in file
    all_attrs = file_attrs.copy()
    all_attrs["platform_name"] = file_attrs["platform"].replace("MTI", "MTG-I")
    all_attrs["source_name"] = file_attrs["data_source"].lower()
    all_attrs["data_provider"] = "EUMETSAT"
    date_time_position = datetime.strptime(
        all_attrs["date_time_position"], "%Y%m%d%H%M%S"
    )
    if "TRAIL" in all_attrs["mtg_name"]:
        # Time coverage start/end will differ across each individual granule that
        # comprises a full scan. So instead of setting the start/end datetime to the
        # time coverage start/end attr in the file metadata, use the date_time_position
        # attr, which is consistent across all files.
        all_attrs["start_datetime"] = date_time_position
        all_attrs["actual_start_datetime"] = datetime.strptime(
            file_attrs["time_coverage_start"], "%Y%m%d%H%M%S"
        )
        all_attrs["end_datetime"] = datetime.strptime(
            file_attrs["time_coverage_end"], "%Y%m%d%H%M%S"
        )
    else:
        all_attrs["start_datetime"] = date_time_position
        all_attrs["end_datetime"] = date_time_position
    all_attrs["scan_datetime"] = date_time_position
    all_attrs["interpolation_radius_of_influence"] = 50 * 1000
    all_attrs["area_definition"] = area_def
    if area_def:
        area_id = area_def.area_id
    else:
        area_id = None
    all_attrs["area_id"] = area_id
    return all_attrs


def load_metadata_single_file(fci_file, group=None, variable=None):
    """Load metadata from an FCI file.

    Parameters
    ----------
    fci_file : str
        Full file path to file
    group : str, optional
        Name of group to load attributes, by default None
    variable : _type_, optional
        Name of variable to load attributes, by default None

    Returns
    -------
    dict
        All metadata attrs for group/variable
    """
    LOG.info("Grabbing metadata for %s", Path(fci_file).name)
    all_meta_attrs = {}
    if not group:
        group = ""
    load_group = []
    for grp in group.split("/"):
        with xarray.open_dataset(fci_file, group="/".join(load_group)) as df:
            meta = df.attrs
        for key, val in meta.items():
            if key not in all_meta_attrs:
                all_meta_attrs[key] = val
        load_group.append(grp)
    if variable and variable in meta:
        for key, val in meta[variable].attrs:
            if key not in all_meta_attrs:
                all_meta_attrs[key] = val
    return all_meta_attrs


def get_ncattrs(fci_file, group=""):
    """Lightweight function to quickly return the attrs for a group in a netCDF file."""
    with ncdf.Dataset(fci_file, "r") as ds:
        attrs = {x: ds[group].getncattr(x) for x in ds[group].ncattrs()}
    return attrs


def avail_chans_in_file(fci_file):
    """Return list of available channels/variables in file.

    Parameters
    ----------
    fci_file : str
        Full file path to FCI file

    Returns
    -------
    list
        list of strings of available variables
    """
    with xarray.open_dataset(fci_file) as df:
        meta = df.l1c_channels_present.data
    return meta.tolist()


def final_calibrated_variable_name(ncvar_name, chan_map=CHAN_MAP):
    """Determine the final name of dataset variable that should be used in GeoIPS.

    Parameters
    ----------
    ncvar_name : str
        Name of variable in source file
    chan_map : dict, optional
        Mapped band name used by GeoIPS, by default CHAN_MAP

    Returns
    -------
    str
        Variable name used in GeoIPS (e.g. B14BT)
    """
    if any([x in ncvar_name for x in ["vis_", "nir_"]]):
        cal_name = "Ref"
    else:
        cal_name = "BT"
    return chan_map[ncvar_name] + cal_name


def scene_to_xarray(fci_files, geoips_chans, metadata=None, area_def=None):
    """Load files into satpy.Scene object then convert to xarray.Dataset.

    Parameters
    ----------
    fci_files : list
        List of FCI netCDF files on disk
    geoips_chans : list
        Channels/variables denoted in GeoIPS nomenclature to load into memory, where
        'GeoIPS nomenclature' is the common variable mapping used in our geostationary
        readers. See 'BAND_MAP' variable for an example of these channels. Allows us
        to read in radiances, reflectances, or brightness temperatures even though the
        variable in the dataset itself (such as 'vis_04') does not provide that
        information.
    metadata : dict, optional
        Standard geoips attributes, by default None
    area_def : pyresample.AreaDefinition, optional
        Crop data to specific area definition, by default None

    Returns
    -------
    xarray.Dataset
        Calibrated FCI MTG data
    """
    scn = Scene(filenames=fci_files, reader="fci_l1c_nc")
    for gchan in geoips_chans:
        rchan = BAND_MAP.get(gchan)
        if rchan is None:
            # Most likely encountered a geolocation variable which will be computed
            # later
            continue
        if "Rad" in gchan:
            calibration = "radiance"
        elif "Ref" in gchan:
            calibration = "reflectance"
        elif "BT" in gchan:
            calibration = "brightness_temperature"
        else:
            # Star denotes the highest calibration, which is BT I think.
            calibration = "*"
        scn.load([rchan], calibration=calibration)
        # We might be deriving data from the same variable multiple times
        # I.e. B01Rad and B01Ref. Rename these data queries to the GeoIPS variable name
        # so we don't have conflicts.
        scn[gchan] = scn[rchan].copy(deep=True)
        # Delete the satpy named variable
        del scn[rchan]

    # Resample all variables in this scene to the finest resolution area definition
    # found. scn.resample() defaults to scn.finest_area()
    resampled_scn = scn.resample(resampler="native")
    LOG.interactive("Resampled scene to finest resolution.")
    if area_def:
        resampled_scn = resampled_scn.crop(area=area_def)

    xr = resampled_scn.to_xarray_dataset()
    lons, lats = resampled_scn.finest_area().get_lonlats()
    # if area_def:
    #     scn = scn.crop(area=area_def)

    # xr = scn.to_xarray_dataset()
    # lons, lats = scn.finest_area().get_lonlats()
    # Mask inf values with -999.9 for compatibility with get_geolocation func
    lons[np.isinf(lons)] = -999.9
    lats[np.isinf(lats)] = -999.9
    geo_meta = get_geolocation_metadata(xr.attrs)
    gvars = get_geolocation(xr.start_time, geo_meta, lats, lons, geo_meta["BADVALS"])
    xr["latitude"] = (("y", "x"), lats)
    xr["longitude"] = (("y", "x"), lons)
    LOG.debug("Added latitude/longitude to xarray dataset")
    for gkey, gvar in gvars.items():
        LOG.debug("Adding %s to xarray dataset", gkey)
        xr[gkey] = (("y", "x"), gvar)
    # Use a set here because gvars could include the same variables as geoips_chans
    # For example, we request satellite and solar zenith angle in our geoips_chans,
    # which is already included in gvars
    for dkey in set(geoips_chans + list(gvars)):
        # Flip data arrays upside down - otherwise image will be upside down in
        # unprojected output
        LOG.debug("Adding %s to xarray dataset", dkey)
        LOG.debug("Flipping data for %s", dkey)
        xr[dkey].data = np.flipud(xr[dkey].compute().data)
        # Convert from wavenumber to wavelength
        if dkey.endswith("Rad"):
            LOG.info("Converting wavenumber to wavelength")
            xr[dkey].data = xr[dkey].data * RADIANCE_CONVERSION_COEFFICIENTS[dkey]

    # Temporarily convert reflectances from % ranging from 0-100 to 0-1
    # This is required until gamma correction can handle reflecances from 0-100
    # This will happen when all readers return reflectance ranging from 0-100
    for dkey in list(xr.variables.keys()):
        if "Ref" in dkey:
            LOG.debug("Converting Ref range to 0-1")
            xr[dkey] /= 100
    # Add all standard GeoIPS attrs, if not already present
    if not metadata:
        metadata = get_standard_geoips_attrs(
            load_metadata_single_file(fci_files[0], group="data"), area_def
        )
        metadata["source_file_names"] = fci_files
    for smeta, sval in metadata.items():
        if smeta not in xr.attrs:
            xr.attrs[smeta] = sval
    return xr


def convert_radiance_to_bt(variable_dataset, line_inds=None, sample_inds=None):
    """Convert effective radiance to effective brightness tempearture.

    The effective brightness temperature of a surface is the temperature of a spatially
    uniform blackbody that emits the equivalent amount of radiant energy as the surface
    within a spectral band characterized by the spectral response function of the
    instrument. Given the band-average spectral radiance per wavenumber Lv, i.e. the
    effective radiance, the effective brightness temperature T_eff can be approximated
    as follows:

    T_eff = (c2 * vc)/(a * ln(1 + (c1 * vc**3) /Lv ) ) - b/a

    The set of coefficients {Vc, A, B}, corresponding to a given spectral response
    function, are found by regression over the required range of temperatures.
    Constants C1 = 2hc2 and C2 = hc/k are radiation constants where c, h, and k are the
    speed of light, Planck, and Boltzmann constant, respectively.

    The variable radiance_to_bt_conversion_coefficient_wavenumber contains the
    wavenumber corresponding to νc.

    The variables radiance_to_bt_conversion_coefficient_a and
    radiance_to_bt_conversion_coefficient_b contain the conversion coefficients a and b
    for IR channels, respectively. They are set to the _FillValue for VNIR channels.

    The variables radiance_to_bt_conversion_constant_c1
    and radiance_to_bt_conversion_constant_c2 contain the
    constants c1 and c2 for IR channels. Note that the values given in the dataset are
    c1=2·1011·hc2 =1.19104282E-05 and c2=100 hc/k=1.43877513 due to unit conversions.
    They are set to the _FillValue for visible and near-infrared channels.

    Parameters
    ----------
    variable_dataset : xarray.Dataset
        The data.<channel>.measured group dataset

    Returns
    -------
    array
        Calibrated brightness temperatures
    """
    LOG.info("Converting effective radiance to brightness temperature")
    a = np.mean(variable_dataset.radiance_to_bt_conversion_coefficient_a)
    b = np.mean(variable_dataset.radiance_to_bt_conversion_coefficient_b)
    c1 = np.mean(variable_dataset.radiance_to_bt_conversion_constant_c1)
    c2 = np.mean(variable_dataset.radiance_to_bt_conversion_constant_c2)
    vc = np.mean(variable_dataset.radiance_to_bt_conversion_coefficient_wavenumber)
    Lv = variable_dataset["effective_radiance"]
    if line_inds is not None:
        # Updated indexing for NumPy 2.3 compatibility:
        # Old code `Lv[line_inds, sample_inds]` fails with 2D indexers.
        # Use xarray's `.isel()` with labeled DataArrays to preserve elementwise
        # indexing and maintain xarray structure (dims, coordinates, and methods)
        # under NumPy 2.3+.
        Lv = Lv.isel(
            y=xarray.DataArray(line_inds, dims=("y", "x")),
            x=xarray.DataArray(sample_inds, dims=("y", "x")),
        )
    t_eff = (c2 * vc) / (a * np.log(1 + (c1 * vc**3) / Lv)) - b / a
    return t_eff


def compute_sun_earth_distance(start_datetime, end_datetime):
    """Compute the sun_earth_distance."""
    middle_time_diff = (start_datetime - end_datetime).total_seconds() / 2
    utc_date = start_datetime + timedelta(seconds=middle_time_diff)
    sun_earth_distance = sun_earth_distance_correction(utc_date)
    LOG.debug(f"The value sun_earth_distance is set to {sun_earth_distance} AU.")
    return sun_earth_distance


def convert_radiance_to_reflectance(
    variable_dataset,
    dataset_name,
    start_datetime,
    end_datetime,
    solar_zenith_angle,
    line_inds=None,
    sample_inds=None,
    earth_sun_distance_km=None,
):
    """Convert radiance to reflectance.

    Parameters
    ----------
    variable_dataset : xarray.Dataset
        Dataset holding all variables from FCI group
    start_datetime : datetime.datetime
        Start time of data
    end_datetime : datetime.datetime
        End time of data

    Returns
    -------
    array
        Calibrated reflectances
    """
    LOG.info("Converting effective radiance to reflectance for %s", dataset_name)
    R = variable_dataset["effective_radiance"]
    # replace any negative radiances with lowest value according to offset and scale
    scaled_low = -R.attrs.get("add_offset", 0) // R.attrs.get("scale_factor", 1) + 1
    R = R.where((~R.notnull()) | (R >= scaled_low), scaled_low)
    # data should already be scaled, but keeping here for reference just in case
    # if dataset_name == "ir_38":
    #     R = xarray.where(
    #         ((2**12 - 1 < R) & (R <= 2**13 - 1)),
    #         (
    #             R * R.attrs.get("warm_scale_factor", 1)
    #             + R.attrs.get("warm_add_offset", 0)
    #         ),
    #         (R * R.attrs.get("scale_factor", 1) + R.attrs.get("add_offset", 0)),
    #     )
    # else:
    #     R = R * R.attrs.get("scale_factor", 1) + R.attrs.get("add_offset", 0)
    if earth_sun_distance_km is None:
        d_t = compute_sun_earth_distance(start_datetime, end_datetime)  # NOQA
    else:
        d_t = earth_sun_distance_km / 149597870.7  # NOQA
    Ir = np.nanmean(variable_dataset.channel_effective_solar_irradiance)  # NOQA
    if line_inds is not None:
        # Updated indexing for NumPy 2.3 compatibility:
        # Old code `Lv[line_inds, sample_inds]` fails with 2D indexers.
        # Use xarray's `.isel()` with labeled DataArrays to preserve elementwise
        # indexing and maintain xarray structure (dims, coordinates, and methods)
        # under NumPy 2.3+.
        R = R.isel(
            y=xarray.DataArray(line_inds, dims=("y", "x")),
            x=xarray.DataArray(sample_inds, dims=("y", "x")),
        )
    brf = np.empty(shape=R.shape)
    pi = np.pi  # NOQA
    cos_sol_zen_ang = np.cos(np.deg2rad(solar_zenith_angle))  # NOQA
    ne.evaluate("(pi * R * d_t**2) / (Ir * cos_sol_zen_ang)", out=brf)
    brf_xarr = xarray.DataArray(
        brf,
        coords={"y": R.coords["y"], "x": R.coords["x"]},
    )
    # log(mean)/log(0.5) for mean rng 0 to 1
    gamma_correction = np.log(np.nanmean(brf)) / np.log(0.5)
    LOG.debug("Expected gamma correction value: %s", gamma_correction)
    return brf_xarr


def read_with_xarray(
    fci_files,
    geoips_chans,
    metadata=None,
    area_def=None,
    geolocation_cache_backend="zarr",
    chunk_size=None,
    cache_solar_angles=False,
    geolocation_only=False,
    precise_earth_sun_distance=True,
    resource_tracker=None,
    satellite_zenith_angle_cutoff=None,
):
    """Load files with xarray and calibrate.

    Parameters
    ----------
    fci_files : list
        List of FCI netCDF files on disk
    geoips_chans : list
        Channels/variables denoted in GeoIPS nomenclature to load into memory, where
        'GeoIPS nomenclature' is the common variable mapping used in our geostationary
        readers. See 'BAND_MAP' variable for an example of these channels. Allows us
        to read in radiances, reflectances, or brightness temperatures even though the
        variable in the dataset itself (such as 'vis_04') does not provide that
        information.
    metadata : dict, optional
        Standard geoips attributes, by default None
    area_def : pyresample.AreaDefinition, optional
        Crop data to specific area definition, by default None
    satellite_zenith_angle_cutoff: cutoff for satellite zenith angle in degrees
        * Defaults to None

    Returns
    -------
    xarray.Dataset
        Calibrated FCI MTG data
    """
    dset = xarray.Dataset()

    for gchan in geoips_chans:
        rchan = BAND_MAP.get(gchan)
        if rchan is None:
            # Most likely encountered a geolocation variable which will be computed
            # later
            continue
        if resource_tracker:
            key = f"READ CHAN: fci_netcdf_chan_{rchan}"
            if area_def:
                key += f"_{area_def.area_id}"
            resource_tracker.track_resource_usage(
                logstr="MEMUSG",
                verbose=False,
                key=key,
                show_log=False,
            )
        with xarray.open_mfdataset(fci_files, group=f"/data/{rchan}/measured") as df:
            if "latitude" not in dset:
                # Add num_lines and num_samples to metadata
                metadata["num_lines"] = df.y.size
                metadata["num_samples"] = df.x.size
                geolocation_metadata = get_geolocation_metadata(metadata)
                fldk_lats, fldk_lons = get_latitude_longitude(
                    geolocation_metadata,
                    df.x.data,
                    df.y.data,
                    BADVALS,
                    geolocation_cache_backend=geolocation_cache_backend,
                    chunk_size=chunk_size,
                    resource_tracker=resource_tracker,
                )
                gvars = get_geolocation(
                    metadata["start_datetime"],
                    geolocation_metadata,
                    fldk_lats,
                    fldk_lons,
                    BADVALS,
                    area_def,
                    geolocation_cache_backend=geolocation_cache_backend,
                    cache_solar_angles=cache_solar_angles,
                    chunk_size=chunk_size,
                    resource_tracker=resource_tracker,
                )
                for gkey, gvar in gvars.items():
                    if gkey in ["Lines", "Samples"]:
                        continue
                    LOG.debug("Adding %s to xarray dataset", gkey)
                    dset[gkey] = (("y", "x"), gvar)
            if geolocation_only:
                continue
            if "Rad" in gchan:
                calibration_type = "Radiance"
                # Convert from wavenumber to wavelength
                eff_rad = df["effective_radiance"][...]
                if "Lines" in gvars and "Samples" in gvars:
                    eff_rad = eff_rad[gvars["Lines"], gvars["Samples"]]
                calibrated = eff_rad * RADIANCE_CONVERSION_COEFFICIENTS[gchan]
            elif "Ref" in gchan:
                calibration_type = "Reflectance"
                calibrated = convert_radiance_to_reflectance(
                    df,
                    dataset_name=rchan,
                    start_datetime=metadata["start_datetime"],
                    end_datetime=metadata["end_datetime"],
                    solar_zenith_angle=dset["solar_zenith_angle"],
                    line_inds=gvars.get("Lines"),
                    sample_inds=gvars.get("Samples"),
                )
            elif "BT" in gchan:
                calibration_type = "Brightness Temperature"
                calibrated = convert_radiance_to_bt(
                    df,
                    line_inds=gvars.get("Lines"),
                    sample_inds=gvars.get("Samples"),
                )
            else:
                raise ValueError(f"Unknown data type to calibrate: {gchan}")
        try:
            LOG.info("Masking data off disk")
            sat_zenith = gvars["satellite_zenith_angle"]
            masked = np.ma.array(calibrated, mask=sat_zenith.mask)
            calibrated.data = masked
            if satellite_zenith_angle_cutoff:
                LOG.info(
                    "Masking data greater than %s degrees sat zenith angle",
                    satellite_zenith_angle_cutoff,
                )
                masked = np.ma.masked_where(
                    gvars["satellite_zenith_angle"] > satellite_zenith_angle_cutoff,
                    calibrated,
                )
                calibrated.data = masked
        except KeyError:
            LOG.info("Did not find satellite zenith angle to mask for %s", gchan)
            pass

        LOG.info("Min %s: %s", calibration_type, float(np.nanmin(calibrated)))
        LOG.info("Max %s: %s", calibration_type, float(np.nanmax(calibrated)))
        LOG.debug("Adding %s to xarray dataset", gchan)
        dset[gchan] = calibrated
        if resource_tracker:
            resource_tracker.track_resource_usage(
                logstr="MEMUSG",
                verbose=False,
                key=key,
                show_log=False,
            )

    if geolocation_only:
        return dset
    # Use a set here because gvars could include the same variables as geoips_chans
    # For example, we request satellite and solar zenith angle in our geoips_chans,
    # which is already included in gvars
    for dkey in set(geoips_chans + list(gvars)):
        # With numpy 2.3 upgrade, sometimes Lines and Samples will exist in
        # the gvars list, but not exist in the dset.
        if dkey not in dset:
            LOG.debug("%s is not in dset", dkey)
            continue
        # Flip data arrays upside down - otherwise image will be upside down in
        # unprojected output
        LOG.debug("Adding %s to xarray dataset", dkey)
        LOG.debug("Flipping data for %s", dkey)
        dset[dkey].data = np.flipud(dset[dkey].data)

    # Add all standard GeoIPS attrs, if not already present
    for smeta, sval in metadata.items():
        if smeta not in dset.attrs:
            dset.attrs[smeta] = sval
    return dset


def read_fci_netcdf_files(
    fci_files,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    use_satpy=False,
    geolocation_cache_backend="zarr",
    cache_chunk_size=None,
    cache_solar_angles=False,
    geolocation_only=False,
    resource_tracker=None,
    satellite_zenith_angle_cutoff=None,
):
    """Read and calibrate data using satpy's fci_l1c_nc reader.

    Parameters
    ----------
    fci_files : list
        list of files for a given scan time
    metadata_only : bool, optional
        Only return metadata, by default False
    chans : list, optional
        List of channels/variables to load, by default None
    area_def : pyresample.AreaDefinition, optional
        Read in data for specific area definition, by default None
    self_register : bool, optional
        Self register data to given resolution, by default False
        (currently unsupported)
    satellite_zenith_angle_cutoff: cutoff for satellite zenith angle in degrees
        * Defaults to None

    Returns
    -------
    dict
        dictionary of xarray datasets
    """
    if area_def and self_register:
        raise ValueError(
            "sector_definition and self_register are mutually exclusive keywords"
        )
    avail_vars = avail_chans_in_file(fci_files[0])
    if not chans:
        # If chans is None, set chans to the GeoIPS band names for each variable found
        # in the FCI files.
        chans = [CHAN_MAP[x] for x in avail_vars]
    ingested = {}
    for fci_file in fci_files:
        if "TRAIL" in fci_file:
            # Use the TRAIL file if possible to
            standard_meta = get_standard_geoips_attrs(
                load_metadata_single_file(fci_file, group="data"), area_def
            )
            break
    else:
        standard_meta = get_standard_geoips_attrs(
            load_metadata_single_file(fci_files[0], group="data"), area_def
        )
    standard_meta["source_file_names"] = fci_files
    if "METADATA" not in ingested:
        ingested["METADATA"] = xarray.Dataset(attrs=standard_meta)
    if metadata_only:
        return ingested
    # Add projection information to top-level metadata, only available in actual data
    # files.
    LOG.info("Getting projection information from %s", Path(fci_files[0]).name)
    with ncdf.Dataset(fci_files[0], "r") as ds:
        plat_group = ds["state/platform"]
        proj_group = ds["data/mtg_geos_projection"]
        for var in plat_group.variables:
            try:
                standard_meta[f"mean_{var}"] = np.mean(plat_group[var][:])
            except AttributeError:
                standard_meta[var] = plat_group[var][:]
        standard_meta["mtg_geos_projection"] = {
            x: proj_group.getncattr(x) for x in proj_group.ncattrs()
        }

    # Loading all chans with satpy will not maintain native resolution for each chan.
    # This is probably OK for now when self-registering (makes things easier as satpy
    # returns all data at the same size of the lowest available resolution)
    if self_register:
        adname = "FULL_DISK"
        if use_satpy:
            ingested[adname] = scene_to_xarray(fci_files, chans, metadata=standard_meta)
        else:
            ingested[adname] = read_with_xarray(
                fci_files,
                chans,
                metadata=standard_meta,
                geolocation_cache_backend=geolocation_cache_backend,
                chunk_size=cache_chunk_size,
                cache_solar_angles=cache_solar_angles,
                geolocation_only=geolocation_only,
                satellite_zenith_angle_cutoff=satellite_zenith_angle_cutoff,
            )
    # Otherwise read in chans with similar resolutions to maintain resolution
    else:
        for res, rchans in DATASET_INFO.items():
            # Find the indices of variables that correspond to the current resolution
            read_idxs = np.array([BAND_MAP.get(x, x) in rchans for x in chans])
            if not any(read_idxs) and not metadata_only:
                LOG.info("No channels to read at %s resolution", res)
                continue
            # Use those indices to retrieve the correct 'geoips' variables that we'll
            # need to read.  Cast np.str to str.
            read_geoips_chans = [str(x) for x in np.array(chans)[read_idxs]]
            LOG.info(
                "Reading in following chans: %s: %s", res, " ".join(read_geoips_chans)
            )
            # Ingest the returned xarray dataset to the current resolution of the
            # xarray_dictionary
            if use_satpy:
                ingested[res] = scene_to_xarray(
                    fci_files,
                    read_geoips_chans,
                    metadata=standard_meta,
                    area_def=area_def,
                )
            else:
                ingested[res] = read_with_xarray(
                    fci_files,
                    read_geoips_chans,
                    metadata=standard_meta,
                    area_def=area_def,
                    geolocation_cache_backend=geolocation_cache_backend,
                    chunk_size=cache_chunk_size,
                    cache_solar_angles=cache_solar_angles,
                    geolocation_only=geolocation_only,
                    resource_tracker=resource_tracker,
                    satellite_zenith_angle_cutoff=satellite_zenith_angle_cutoff,
                )
    return ingested


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    use_satpy=False,
    geolocation_cache_backend=gpaths["GEOIPS_GEOLOCATION_CACHE_BACKEND"],
    cache_chunk_size=None,
    cache_solar_angles=False,
    geolocation_only=False,
    resource_tracker=None,
    satellite_zenith_angle_cutoff=None,
):
    """Read and calibrate FCI data.

    Parameters
    ----------
    fci_files : list
        list of files for a given scan time
    metadata_only : bool, optional
        Only return metadata, by default False
    chans : list, optional
        List of channels/variables to load, by default None
    area_def : pyresample area definition, optional
        Read in data for specific area definition, by default None
        (currently unsupported)
    self_register : bool, optional
        Self register data to given resolution, by default False
        (currently unsupported)
    satellite_zenith_angle_cutoff: cutoff for satellite zenith angle in degrees
        * Defaults to None

    Returns
    -------
    dict
        dictionary of xarray datasets
    """
    if metadata_only:
        # If metadata_only is True, only pass one file to the reader. Otherwise the
        # readers.read_data_to_xarray_dict method will crack open every FCI file,
        # which more or less has the same metadata information - the only field that
        # changes to my knowledge is the start and end time. The TRAIL file has the
        # overall start and end time across the files, so use that if we can. Otherwise
        # take the first file in the list.
        trail_files = []
        for fci_file in fnames:
            if "TRAIL" in fci_file:
                trail_files.append(fci_file)
        if trail_files:
            fnames = trail_files
    fci_data = readers.read_data_to_xarray_dict(
        fnames,
        read_fci_netcdf_files,
        metadata_only,
        chans,
        area_def,
        self_register,
        use_satpy=use_satpy,
        geolocation_cache_backend=geolocation_cache_backend,
        cache_chunk_size=cache_chunk_size,
        cache_solar_angles=cache_solar_angles,
        geolocation_only=geolocation_only,
        satellite_zenith_angle_cutoff=satellite_zenith_angle_cutoff,
    )

    return fci_data



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Ancillary module containing test dataset information.

Dictionary mapping of GeoIPS Test Datasets.

Mapping goes {"test_dataset_name": "test_dataset_url"}
"""

from pathlib import Path
import yaml
from typing import Dict

# This is not a plugin module
interface = None


def load_test_dataset_urls() -> Dict[str, str]:
    """
    Load test dataset URLs from the YAML configuration file.

    Returns
    -------
    Dict[str, str]
        Dictionary mapping test dataset names to their URLs
    """
    current_file_path = Path(__file__).resolve()
    url_file_path = current_file_path.parent / "test-data-urls.yaml"

    with open(url_file_path, "r") as file:
        urls = yaml.safe_load(file)

    return urls["test_data_urls"]


# Initialize the test dataset dictionary
test_dataset_dict = load_test_dataset_urls()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit tests for geoips.utils.cache_files."""

import os
import shutil
import yaml
import json
import tempfile

from geoips.utils.cache_files import (
    source_modified,
    create_cached_json_from_yaml,
    get_cached_json,
)


def test_source_modified():
    """Test the source_modified function.

    Create a temporary source and destination file, then test that source_modified
    returns True if the source file is newer than the destination file, and False
    otherwise. This will also test the case where the destination file does not exist.
    """
    tmp_dir = tempfile.mkdtemp()
    cache_dir = tempfile.mkdtemp()

    # Create a temporary source and destination file
    source = f"{tmp_dir}/geoips_test_source.txt"
    dest = f"{cache_dir}/test_dest.txt"

    with open(source, "w") as f:
        f.write("This is a test.")

    # Test that source_modified returns True if the destination file does not exist.
    assert source_modified(source, dest)

    os.makedirs(cache_dir, exist_ok=True)
    with open(dest, "w") as f:
        f.write("This is a test.")

    # Test that source_modified returns True if the source file is newer
    os.utime(source, (os.path.getatime(source), os.path.getmtime(source) + 1))
    assert source_modified(source, dest)

    # Test that source_modified returns False if the destination file is newer
    os.utime(dest, (os.path.getatime(dest), os.path.getmtime(dest) + 1))
    assert not source_modified(source, dest)

    # Clean up
    os.remove(source)
    os.remove(dest)
    os.rmdir(cache_dir)
    os.rmdir(tmp_dir)


def test_create_cached_json_from_yaml():
    """Test the create_cached_json_from_yaml function.

    This will create a temporary YAML file, then call create_cached_json_from_yaml on
    it. This will read the YAML file write the data to a JSON file in `cache_dir`. The
    test just ensures that the JSON file was created and contains the correct data.
    """
    tmp_dir = tempfile.mkdtemp()
    cache_dir = tempfile.mkdtemp()

    # Create a temporary YAML file
    yaml_file = f"{tmp_dir}/test.yaml"
    with open(yaml_file, "w") as f:
        yaml.dump({"key": "value"}, f)

    # Create the cached JSON file
    json_file = create_cached_json_from_yaml(yaml_file, cache_dir=cache_dir)

    # Check that the JSON file was created and contains the correct data
    assert os.path.exists(json_file)
    with open(json_file, "r") as f:
        data = json.load(f)
        assert data == {"key": "value"}

    # Clean up
    os.remove(yaml_file)
    os.remove(json_file)
    os.rmdir(cache_dir)
    os.rmdir(tmp_dir)


def test_get_cached_json():
    """Test the get_cached_json function.

    Write a YAML file, then call get_cached_json on it. This will read the YAML file,
    write it to a JSON file in `cache_dir`, read the JSON file and return the data, then
    compare the new data to the original YAML data.
    """
    tmp_dir = tempfile.mkdtemp()
    cache_dir = tempfile.mkdtemp()

    # Create a temporary YAML file
    yaml_file = f"{tmp_dir}/test.yaml"
    with open(yaml_file, "w") as f:
        yaml.dump({"key": "value"}, f)

    # Get the cached JSON file
    data = get_cached_json(yaml_file, cache_dir=cache_dir)
    assert data == {"key": "value"}

    # Clean up
    os.remove(yaml_file)
    shutil.rmtree(cache_dir)
    os.rmdir(tmp_dir)
